//
// Portable Integrity Protection (PIP) System -
// Copyright (C) 2015 Secure Systems Laboratory, Stony Brook University
//
// This file is part of Portable Integrity Protection (PIP) System.
//
// Portable Integrity Protection (PIP) System is free software: you can
// redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Portable Integrity Protection (PIP) System is distributed in the hope that it
// will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Portable Integrity Protection (PIP) System.  If not, see
// <http://www.gnu.org/licenses/>.
//

#include <bitset>

#include "ppi_bufferManager.h"
#include "ppi_debug.h"

/* Custom buffer pool for allocating memory mainly for pathnames */

//Mutex for locking the pool
HANDLE ppi_bufferPool_bit_mtx;

int ppi_bufferPool_curPos = 0;
BOOL ppi_bm_initialized = FALSE;

std::bitset<PPI_BUFFERPOOL_SIZE> ppi_bufferPool_bit;
char ppi_bufferPool[PPI_BUFFERPOOL_SIZE * PPI_BUFFERPOOL_PER_ENTRY_SIZE];

#ifdef PPI_BM_BUFFER_OVERFLOW_DETECTION
//For detecting buffer overflow from ppi_bufferPool
unsigned int guard[10240];

#endif

int ppi_bm_maxUsage = 0;
int ppi_bm_curUsage = 0;

#define ppi_bufferPool_curPos_inc \
	do { \
		ppi_bufferPool_curPos++; \
		ppi_bufferPool_curPos %= PPI_BUFFERPOOL_SIZE; \
	} while (0)



void ppi_bm_init() {
	ppi_bufferPool_bit_mtx = CreateMutex(NULL, FALSE, NULL);
	if (ppi_bufferPool_bit_mtx == NULL) {
		PPI_CRITICAL("Failed to initialize buffer manager");
		return;
	}
	ppi_bm_initialized = TRUE;

#ifdef PPI_BM_BUFFER_OVERFLOW_DETECTION	
	guard[0] = 0xffffffff;
	guard[1] = 0xffffffff;
#endif

}


#ifdef PPI_BM_BUFFER_OVERFLOW_DETECTION	
void *ppi_bm_getProtected() {
	return guard;
}

bool ppi_bm_hasCorrupted() {
	return (guard[0] != 0xffffffff || guard[1] != 0xffffffff);
}
#endif

void *ppi_bm_malloc(int size) {
	
	if (!ppi_bm_initialized)
		ppi_bm_init();

	char *result = NULL;

	if (size > PPI_BUFFERPOOL_PER_ENTRY_SIZE)
		return malloc(size);

	if (ppi_bufferPool_bit.all()) {
		PPI_CRITICAL("All buffers are used!");
		return malloc(size);
	}

	DWORD dwWaitResult = WaitForSingleObject(ppi_bufferPool_bit_mtx, 1000);
	if (dwWaitResult != WAIT_OBJECT_0) {
		void *rv = malloc(size);
		PPI_CRITICAL("Failed to wait for the bm mutex: current usage: %d. Allocating %x, result: %x, error: %x", ppi_bm_curUsage, rv, dwWaitResult, GetLastError());
		return rv;
	}

	int searchStartPos = ppi_bufferPool_curPos;

	ppi_bufferPool_curPos_inc;
	while (ppi_bufferPool_curPos != searchStartPos) {
		if (!ppi_bufferPool_bit[ppi_bufferPool_curPos])
			break;
		ppi_bufferPool_curPos_inc;
	}

	if (ppi_bufferPool_bit[ppi_bufferPool_curPos]) {
		PPI_CRITICAL("Something is not correct!!!!!!");
		ReleaseMutex(ppi_bufferPool_bit_mtx);
		return malloc(size);
	}

	ppi_bufferPool_bit.set(ppi_bufferPool_curPos);
	ppi_bm_curUsage++;
	
	
	result = ppi_bufferPool_curPos*PPI_BUFFERPOOL_PER_ENTRY_SIZE + ppi_bufferPool;
	//PPI_CRITICAL("Buffer Idx is %d: %x", ppi_bufferPool_curPos, result);
	//PPI_CRITICAL("using entry %d, address %Xh, from %Xh to %Xh", ppi_bufferPool_curPos, result, ppi_bufferPool, ppi_bufferPool + PPI_BUFFERPOOL_PER_ENTRY_SIZE * PPI_BUFFERPOOL_SIZE);

	
	ReleaseMutex(ppi_bufferPool_bit_mtx);

	if (ppi_bm_curUsage > ppi_bm_maxUsage) {
		ppi_bm_maxUsage = ppi_bm_curUsage;
		PPI_INFO("Max buffer usage is %d", ppi_bm_maxUsage);	
	}

	return result;

}

BOOL ppi_bm_isbuffer(const void *buf) {
	return !(buf < ppi_bufferPool || (buf >= (ppi_bufferPool + PPI_BUFFERPOOL_PER_ENTRY_SIZE * PPI_BUFFERPOOL_SIZE)));
}



void ppi_bm_free(void *tofree) {

#ifdef PPI_BM_BUFFER_OVERFLOW_DETECTION	
	if (ppi_bm_hasCorrupted()) {
		PPI_CRITICAL("Guard is corrupted!!!: %x", guard);
	}
#endif

	if (tofree < ppi_bufferPool || (tofree >= (ppi_bufferPool + PPI_BUFFERPOOL_PER_ENTRY_SIZE * PPI_BUFFERPOOL_SIZE))) {
		PPI_CRITICAL("calling free on %x", tofree);
		free(tofree);
		return;
	}
	
	int entry = ((unsigned)tofree - (unsigned)ppi_bufferPool)/PPI_BUFFERPOOL_PER_ENTRY_SIZE;

	if (!ppi_bufferPool_bit.test(entry)) {
		PPI_CRITICAL("Trying to free an unused entry!!!!!");
		goto out;
	}

	DWORD dwWaitResult = WaitForSingleObject(ppi_bufferPool_bit_mtx, 1000);
	if (dwWaitResult != WAIT_OBJECT_0) {
		PPI_CRITICAL("Failed to wait for the bm mutex...: Result: %x, error: %d releasing the pool %x unconditionally???", dwWaitResult, GetLastError(), tofree);
		return;
	}

	ppi_bufferPool_bit.reset(entry);
	ppi_bm_curUsage--;

	ReleaseMutex(ppi_bufferPool_bit_mtx);

out:
	return;
}

int ppi_bm_copy(const void *src, void *dst) {
	if (ppi_bm_isbuffer(src) && ppi_bm_isbuffer(dst)) {
		memcpy(dst, (void *)src, PPI_BUFFERPOOL_ENTRY_SIZE);
		return 0;
	}
	return -1;
}
