//
// Portable Integrity Protection (PIP) System -
// Copyright (C) 2015 Secure Systems Laboratory, Stony Brook University
//
// This file is part of Portable Integrity Protection (PIP) System.
//
// Portable Integrity Protection (PIP) System is free software: you can
// redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Portable Integrity Protection (PIP) System is distributed in the hope that it
// will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Portable Integrity Protection (PIP) System.  If not, see
// <http://www.gnu.org/licenses/>.
//

#include "ppi_helperConnection.h"

#define WIN32_LEAN_AND_MEAN             // Exclude rarely-used stuff from Windows headers
// Windows Header Files:
#include <windows.h>
#include <winternl.h>



#include <windows.h>
#include <ws2tcpip.h>
#include <stdlib.h>
#include <stdio.h>

#include "ppi_debug.h"
#include "ppi_config.h"
#include "ppi_packet.h"
#include "ppi_utils.h"

#pragma comment (lib, "Ws2_32.lib")

/* Setting up connection to helper */


SOCKET ppi_getConnection2Helper(void);
SOCKET connection2helper = INVALID_SOCKET;

HANDLE ppi_helperConnection_mtx = NULL;


bool WinsockInitialized() {
	SOCKET s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (s == INVALID_SOCKET && WSAGetLastError() == WSANOTINITIALISED) {
		return false;
	}
	return true;
}

void ppi_helper_init() {

    WSADATA wsaData;
    connection2helper = INVALID_SOCKET;

    int iResult;
	
	if (ppi_helperConnection_mtx == NULL) {
		ppi_helperConnection_mtx = CreateMutex(NULL, FALSE, NULL);
		if (ppi_helperConnection_mtx == NULL) {
			PPI_CRITICAL("Failed to initialize helper connection lock");
		}
	}

    // Initialize Winsock
	if (!WinsockInitialized()) {
		iResult = WSAStartup(MAKEWORD(2,2), &wsaData);
	} else
		iResult = 0;
	
    if (iResult != 0)
        PPI_ERROR("WSAStartup failed with error: %d", iResult);
	
	return;
}

BOOL ppi_helper_sendRecvCorrectlyWithLock(struct ppi_pkt *_pkt, struct ppi_pkt *_response){
	int rv = false;
	
	DWORD dwWaitResult = WaitForSingleObject(ppi_helperConnection_mtx, INFINITE);
	if (dwWaitResult != WAIT_OBJECT_0) {		
		PPI_CRITICAL("Failed to wait for the helper mutex");
		if (dwWaitResult == WAIT_FAILED)
			PPI_CRITICAL("Failed with error: GetLastError: %x", GetLastError());
	}

	rv = ((ppi_sendPkt2Helper(_pkt) == 0) && (ppi_recvPktFromHelper(_response) == 0));

	if (ppi_helperConnection_mtx != NULL)
		ReleaseMutex(ppi_helperConnection_mtx);
	
	return rv;
}


void ppi_helper_fin() {
	
	if (connection2helper != INVALID_SOCKET)
		closesocket(connection2helper);
    WSACleanup();
}


int ppi_sendPkt2Helper(struct ppi_pkt *pkt) {
	if (connection2helper != INVALID_SOCKET)
		goto send_data;

	if (ppi_getConnection2Helper() == INVALID_SOCKET)
		return -1;

send_data:
	if (!SEND_PKT_CORRECT(connection2helper, *pkt)) {
		PPI_CRITICAL("Failed to send packet correctly");
		return -1;
	}
	return 0;
}


int ppi_recvPktFromHelper(struct ppi_pkt *pkt) {
	int expectedResponseCMDNo = pkt->l_cmdno;
	if (connection2helper == INVALID_SOCKET) {
		PPI_CRITICAL("Trying to receive from an invalid socket");
		return -1;
	}
	if (!RECV_PKT_CORRECT(connection2helper, *pkt)) {
		PPI_CRITICAL("Failed to receive from helper correctly");
		return -1;
	}
	if (expectedResponseCMDNo != pkt->l_cmdno) {
		PPI_CRITICAL("Incorrect response received from helper: Expected: %d, Received: %d", expectedResponseCMDNo, pkt->l_cmdno);
		return -1;
	}
	return 0;
}

BOOL tried = FALSE;
SOCKET ppi_getConnection2Helper() {
	int iResult;
    struct addrinfo *result = NULL,
                    *ptr = NULL,
                    hints;

	if (!WinsockInitialized())
		ppi_helper_init();

	if (connection2helper != INVALID_SOCKET)
		goto out;
	
	if (tried)
		goto out;

    ZeroMemory( &hints, sizeof(hints) );
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;

    // Resolve the server address and port
    iResult = getaddrinfo("127.0.0.1", DEFAULT_PORT, &hints, &result);
    if ( iResult != 0 ) {
        PPI_ERROR("getaddrinfo failed with error: %d", iResult);
		goto out;
    }

    for(ptr=result; ptr != NULL ;ptr=ptr->ai_next) {

        // Create a SOCKET for connecting to helper
        connection2helper = socket(ptr->ai_family, ptr->ai_socktype, 
            ptr->ai_protocol);
        if (connection2helper == INVALID_SOCKET) {
            PPI_ERROR("socket failed with error: %ld\n", WSAGetLastError());
			goto free_result;
        }

        // Connect to helper.
        iResult = connect(connection2helper, ptr->ai_addr, (int)ptr->ai_addrlen);
        if (iResult == SOCKET_ERROR) {
			PPI_ERROR("Failed to connect to helper");
            closesocket(connection2helper);
            connection2helper = INVALID_SOCKET;
            continue;
        }
        break;
    }

	//tried = TRUE;

free_result:
    freeaddrinfo(result);

out:
	return connection2helper;
}

