//
// Portable Integrity Protection (PIP) System -
// Copyright (C) 2015 Secure Systems Laboratory, Stony Brook University
//
// This file is part of Portable Integrity Protection (PIP) System.
//
// Portable Integrity Protection (PIP) System is free software: you can
// redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Portable Integrity Protection (PIP) System is distributed in the hope that it
// will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Portable Integrity Protection (PIP) System.  If not, see
// <http://www.gnu.org/licenses/>.
//

#include <string.h>
#include <uthash.h>
#include "ppi_helper_handleCacheManager.h"
#include "ppi_debug.h"
#include "ppi_bufferManager.h"

/* Cache the handle at the helper side */

// this is an example of how to do a LRU cache in C using uthash
// http://uthash.sourceforge.net/
// by Jehiah Czebotar 2011 - jehiah@gmail.com
// this code is in the public domain http://unlicense.org/

#define MAX_CACHE_SIZE 100


typedef struct {
	HANDLE processHandle;
	HANDLE processFileHandle;
} record_key_t;

struct CacheEntry {
	record_key_t key;
	HANDLE localFileHandle;
	UT_hash_handle hh;
};
struct CacheEntry *cache = NULL;


//Invalidate the cached handle
void ppi_helper_handleCacheManager_closeHandle(HANDLE processHandle, HANDLE processFileHandle)
{
	record_key_t key;
	struct CacheEntry *entry;
	
	key.processHandle = processHandle;
	key.processFileHandle = processFileHandle;

	HASH_FIND(hh, cache, &key, sizeof(record_key_t), entry);

	if (entry) {
		HASH_DELETE(hh, cache, entry);
		ppi_callOriginal(NtClose, entry->localFileHandle);
		ppi_bm_free(entry);
	}
}

//Check from the cache to see if a handle is cached, or duplicate it when necessary
HANDLE ppi_helper_handleCacheManager_getPeerHandle(HANDLE processHandle, HANDLE processFileHandle)
{
	record_key_t key;
	struct CacheEntry *entry;
	
	key.processHandle = processHandle;
	key.processFileHandle = processFileHandle;

	HASH_FIND(hh, cache, &key, sizeof(record_key_t), entry);

	if (entry) {
		// remove it (so the subsequent add will throw it on the front of the list)
		HASH_DELETE(hh, cache, entry);
		HASH_ADD(hh, cache, key, sizeof(record_key_t), entry);
		return entry->localFileHandle;

	} else {

		// duplicate the handle, insert into lru, remove oldest if too large

		HANDLE localHandle = NULL;
		if (DuplicateHandle(processHandle, processFileHandle, GetCurrentProcess(), &localHandle, NULL, true, DUPLICATE_SAME_ACCESS) != 0) {
			struct CacheEntry *entry, *tmp_entry;
			entry = (struct CacheEntry *)ppi_bm_malloc(sizeof(struct CacheEntry));
			entry->localFileHandle = localHandle;
			entry->key = key;
			HASH_ADD(hh, cache, key, sizeof(record_key_t), entry);
						
			// prune the cache to MAX_CACHE_SIZE
			if (HASH_COUNT(cache) >= MAX_CACHE_SIZE) {
				HASH_ITER(hh, cache, entry, tmp_entry) {
					// prune the first entry (loop is based on insertion order so this deletes the oldest item)
					HASH_DELETE(hh, cache, entry);
					ppi_callOriginal(NtClose, entry->localFileHandle);
					free(entry);
					break;
				}
			}
			return localHandle;
		} else {
			PPI_CRITICAL("Cannot duplicate for rename, handle in peer %Xh: %Xh", processHandle, processFileHandle);
			return NULL;
		}
	}
	return NULL;
}
