//
// Portable Integrity Protection (PIP) System -
// Copyright (C) 2015 Secure Systems Laboratory, Stony Brook University
//
// This file is part of Portable Integrity Protection (PIP) System.
//
// Portable Integrity Protection (PIP) System is free software: you can
// redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Portable Integrity Protection (PIP) System is distributed in the hope that it
// will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Portable Integrity Protection (PIP) System.  If not, see
// <http://www.gnu.org/licenses/>.
//

#include <ntstatus.h>
#define WIN32_NO_STATUS

#define WIN32_LEAN_AND_MEAN             // Exclude rarely-used stuff from Windows headers
// Windows Header Files:
#include <windows.h>
#include <winternl.h>
#include <string>
#include <Sddl.h>
#include <Aclapi.h>
#include <set>


#include "ppi_debug.h"
#include "ppi_utils.h"
#include "ppi_config.h"
#include "ppi_version.h"

#ifdef WIN81
#include <atlbase.h>
#include <urlmon.h>
#include <stdlib.h>
#endif

#include "ppi_principal.h"

#include "ppi_bufferManager.h"
#include "ppi_level.h"

#include <algorithm>

#pragma comment (lib, "Advapi32.lib")

/* Manage level related information */


const wchar_t *untrustedInGroupSID[] = {
	PPI_UNTRUSTED_SIDSTRING,
	PPI_EVERYONE_SIDSTRING,
	PPI_USERS_SIDSTRING,
	NULL};


//Case insensitive
const wchar_t *ppi_benignToReadDirs[] = {
	L"pipe\\cygwin-",
	PPI_ALL_LOG_DIR,
	L"\\DEVICE\\",
	L"BitLocker\\Volume\\",
	
	//For Adobe Reader
	L"C:\\Program Files\\Common Files\\Adobe\\SLCache\\",
	L"C:\\ProgramData\\Adobe\\SLStore\\",

	L"hdaudio#",
	L"pipe\\TSVCPIPE",
	L"pipe\\IPRT-VBoxTrayIPC",
	L"Volume{",
	L"C:\\ProgramData\\Adobe\\Extension Manager CC\\Configuration\\",
	L"ide#cdromvbox_cd-rom",
	L"acpi#pnp0c0a",
	NULL
};

//Case sensitive
const std::wstring ppi_benignToReadFiles[] = {
	L"PIPE\\lsarpc", //System related
	L"VBoxMiniRdrDN",
	L"WMIDataDevice",
	L"PIPE\\srvsvc",
	L"NDISTAPI",
	L"VBoxGuest",
	L"pipe\\winlogonrpc",
	L"PIPE\\wkssvc",
	L"VBoxMiniRdrDN",
	L"PIPE\\ROUTER",
	L"PIPE\\EVENTLOG",
	L"PIPE\\NETLOGON",
	L"PIPE\\samr",
	L"PIPE\\DAV RPC SERVICE",
	L"DBGV",
	L"Global\\DBGV",
	L"C:\\Documents and Settings\\Ubuntu\\Desktop\\malware\\runMe.bat",
	L"\\Device\\Afd\\Endpoint",
	L"HCD0",
	L"HCD1",
	L"\\Device\\Afd\\AsyncConnectHlp",
	L"\\Device\\Termdd",
	L"root#system#0000#{3e227e76-690d-11d2-8161-0000f8775bf1}\\{cd171de3-69e5-11d2-b56d-0000f8754380}&{9b365890-165f-11d0-a195-0020afd156e4}",
	L"\\Device\\{8111FD07-9483-4415-9A63-B423D51C61FB}",
	L"\\Device\\KsecDD",
	L"\\Device\\Null",

	/**** Win 8.1 ****/
	L"\\Device\\DeviceApi\\CMApi",
	L"C:\\Windows\\SYSTEM32\\d3d9.dll",
	L"C:\\Windows\\system32\\d3d9.dll",
	L"pipe\\MsFteWds",
	L"\\Device\\DeviceApi\\Dev\\Query",
	L"\\Device\\DeviceApi\\CMNotify",
	//L"C:\\Windows\\System32\\VBoxTray.exe",
	L"\\Device\\{5954BC2D-C880-4A7A-8935-7B1D84F4D1BE}",
	L"Volume{88219ca8-2a4a-11e4-9717-806e6f6e6963}",
	L"C:\\Windows\\SYSTEM32\\D3D9.DLL",
	L"nul",
	L"\\Device\\{F279E122-DCCC-460F-91DC-7BEE5AB77956}",
	L"C:\\Windows\\System32\\Tasks\\Microsoft\\Windows\\CertificateServicesClient\\UserTask-Roam"
};

const std::set<std::wstring> ppi_benignToReadFilesSet(std::begin(ppi_benignToReadFiles), std::end(ppi_benignToReadFiles));


#define ppi_level_benign2read(_lpFileName) (ppi_benignToReadFilesSet.find(std::wstring(_lpFileName))!= ppi_benignToReadFilesSet.end() || ppi_utils_wstringStartWithArray((_lpFileName), ppi_benignToReadDirs))


#define ASSIGN_OWNER(_path, _owner) override_ownership[_path] =  PPI_PRINCIPAL_GET_PRINCIPAL( _owner);


// Customize specific file's level
void PPI_OVERRIDE_OWNERSHIP_MANAGER::initialize() {

#ifndef PPI_TWO_PRINCIPAL
	//Override the ownership information for specific files
	ASSIGN_OWNER(L"C:\\Program Files\\Microsoft Office\\Office14\\WINWORD.EXE", PPI_MULTIP_MICROSOFT);
	ASSIGN_OWNER(L"C:\\Program Files\\Microsoft Office\\Office14\\POWERPNT.EXE", PPI_MULTIP_MICROSOFT);
#endif
	
	for(const wstring *s = begin(ppi_benignToReadFiles); s < end(ppi_benignToReadFiles); s++) {
		ASSIGN_OWNER(*s, PPI_MULTIP_BENIGN);
	}

}


const PPI_PRINCIPAL *PPI_OVERRIDE_OWNERSHIP_MANAGER::getOverridedOwner(const wchar_t *file) const {

	map<const wstring, const PPI_PRINCIPAL *>::const_iterator it = override_ownership.find(file);
	if (it != override_ownership.end())
		return it->second;

	if (ppi_benignToReadFilesSet.find(std::wstring(file))!= ppi_benignToReadFilesSet.end() || ppi_utils_wstringStartWithArray((file), ppi_benignToReadDirs))
		return PPI_PRINCIPAL_GET_PRINCIPAL(PPI_MULTIP_BENIGN);


	return nullptr;
}



int ppi_level_pSDIsUntrustedAclCheck(const PSECURITY_DESCRIPTOR pSD, LPBOOL lpbIsUntrusted);
int ppi_level_pSDIsUntrustedW(const PSECURITY_DESCRIPTOR pSD, LPCWSTR fileName, LPBOOL lpBpSDIsUntrusted);


int ppi_level_getMSFileIsUntrustedW(LPCWSTR lpFileName, BOOL *bFileIsUntrusted) {
	
	HANDLE hStream;
	char *dbuf = NULL;
	DWORD byteRead;
	int rv = -1;
	wchar_t *buf = NULL;
	
	if (lpFileName == NULL) goto out;

	int origPathLen = wcslen(lpFileName);
	if (origPathLen + wcslen(L":Zone.Identifier") > PPI_PATH_WCHAR_SIZE) goto out;
	
	dbuf = (char *)ppi_bm_malloc(200);
	if (dbuf == NULL) goto out;

	buf = (wchar_t *)ppi_bm_malloc(PPI_PATH_WCHAR_SIZE*sizeof(wchar_t));
	if (buf == NULL) goto out;

	memcpy(buf, lpFileName, origPathLen*sizeof(wchar_t));
	memcpy(buf + origPathLen, L":Zone.Identifier", wcslen(L":Zone.Identifier")*sizeof(wchar_t));
	buf[origPathLen + wcslen(L":Zone.Identifier")] = L'\0';
	
	hStream = CreateFile(buf, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if (hStream == INVALID_HANDLE_VALUE) {
		//PPI_ERROR("Cannot open file to check for integrity: error %ld", GetLastError());
		*bFileIsUntrusted = FALSE;
		rv = 0;
		goto out;
	}
		
	*bFileIsUntrusted = FALSE;
	rv = 0;

	if (ReadFile(hStream, dbuf, 200, &byteRead, NULL)) {
		if (byteRead > 0) {
			dbuf[byteRead] = '\0';
			int zoneID;
			if (sscanf(dbuf, "[ZoneTransfer]\nZoneId=%d", &zoneID) == 1) {
				*bFileIsUntrusted = (zoneID >= 3);
				PPI_ERROR("ZoneID for file %ls is %d", buf, zoneID);
			}
			else {
				PPI_ERROR("Content is not expected: %s", dbuf);
			}
		}
	}
	ppi_callOriginal(NtClose, hStream);
	
out:
	if (buf != NULL)
		ppi_bm_free(buf);
	if (dbuf != NULL)
		ppi_bm_free(dbuf);
	return rv;
}


int ppi_getFileIsUntrustedW(LPCWSTR lpFileName, LPBOOL lpBFileIsUntrusted) {

	int rv = -1;
	//PACL pACL = NULL;
	PSECURITY_DESCRIPTOR pSD = NULL;

	if (wcsncmp(lpFileName, L"\\??\\", 4) == 0)
		lpFileName += 4;
	
	DWORD dwRes = GetNamedSecurityInfoW(lpFileName,
		SE_FILE_OBJECT, DACL_SECURITY_INFORMATION|OWNER_SECURITY_INFORMATION|GROUP_SECURITY_INFORMATION, NULL, NULL, NULL, NULL, &pSD);

	if (ERROR_SUCCESS != dwRes) {
		if ((dwRes == ERROR_FILE_NOT_FOUND) || (dwRes == ERROR_INVALID_PARAMETER)) {
			goto out;
		}
		//PPI_ERROR("Failed to get pSD with file %ls. errno: %d", lpFileName, dwRes);
		goto out;
		
	} else {
	
		rv = ppi_level_pSDIsUntrustedW(pSD, lpFileName, lpBFileIsUntrusted);
	}

out:
	if (pSD != NULL)
		LocalFree((HLOCAL)pSD);
	
	if (lpFileName && (*lpBFileIsUntrusted || rv != 0)) {

		if (ppi_level_benign2read(lpFileName)) {
			*lpBFileIsUntrusted = false;
			rv = 0;
		}
	}
	
	if (lpFileName && rv == 0 && *lpBFileIsUntrusted) {
		DWORD rvt = GetFileAttributes(lpFileName);
		if (rvt != INVALID_FILE_ATTRIBUTES && (rvt & FILE_ATTRIBUTE_DIRECTORY || rvt & FILE_ATTRIBUTE_DEVICE))
			*lpBFileIsUntrusted = false;
	}
	return rv;
}


#define PPI_DEFULT_BUFFER_SIZE PPI_BUFFERPOOL_PER_ENTRY_SIZE

int ppi_getHandleIsUntrustedW(HANDLE handle, LPCWSTR lpFileName, LPBOOL lpBHandleIsUntrusted) {

	bool handleIsUntrusted = true;
	DWORD lenNeeded = PPI_BUFFERPOOL_PER_ENTRY_SIZE;
	BOOL tempRes;
	PSECURITY_DESCRIPTOR pSD = NULL;
	int rv = -1;

	pSD = (PSECURITY_DESCRIPTOR)ppi_bm_malloc(lenNeeded);
	tempRes = GetKernelObjectSecurity(handle, DACL_SECURITY_INFORMATION|GROUP_SECURITY_INFORMATION|OWNER_SECURITY_INFORMATION, NULL, 0, &lenNeeded);
	if (tempRes)
		goto checkpSD;
	ppi_bm_free(pSD);
	if(GetLastError() == ERROR_INSUFFICIENT_BUFFER) {
		pSD = (PSECURITY_DESCRIPTOR)ppi_bm_malloc(lenNeeded);
		if (pSD != NULL) {
			tempRes = GetKernelObjectSecurity(handle, DACL_SECURITY_INFORMATION|GROUP_SECURITY_INFORMATION|OWNER_SECURITY_INFORMATION, pSD, lenNeeded, &lenNeeded);
			if (tempRes) {
checkpSD:
				rv = ppi_level_pSDIsUntrustedW(pSD, lpFileName, lpBHandleIsUntrusted);
			} else if (GetLastError() != ERROR_INVALID_FUNCTION) {
				PPI_ERROR("Failed to perform GetKernelObjectSecurity on file %ls handle: %lu due to unknown reason: %d", lpFileName, handle, GetLastError());
			}
			ppi_bm_free(pSD);
		}
	} else {
		if (GetLastError() == ERROR_ACCESS_DENIED) {
			
			if (lpFileName == NULL)
				goto out;

			DWORD dwRes = GetNamedSecurityInfo(lpFileName, 
				SE_FILE_OBJECT, DACL_SECURITY_INFORMATION|OWNER_SECURITY_INFORMATION|GROUP_SECURITY_INFORMATION, NULL, NULL, NULL, NULL, &pSD);

			if (ERROR_SUCCESS != dwRes) {
				PPI_ERROR("Benign process failed to call GetKernelObjectSecurity and GetNamedSecurityInfo on file %ls handle %lu due to permission reason... Assuming file is benign", lpFileName, handle);
				*lpBHandleIsUntrusted = FALSE;
				rv = 0;
			} else {
				rv = ppi_level_pSDIsUntrustedW(pSD, lpFileName, lpBHandleIsUntrusted);
				LocalFree((HLOCAL)pSD);
			}
			goto out;
		} else if (GetLastError() != ERROR_INVALID_FUNCTION)
			PPI_ERROR("Failed to perform GetKernelObjectSecurity on file %ls handle: %lu due to unknown reason: %d", lpFileName, handle, GetLastError());
	}

out:
	if (lpFileName && (*lpBHandleIsUntrusted || rv != 0)) {

		if (ppi_level_benign2read(lpFileName)) {
			*lpBHandleIsUntrusted = FALSE;
			rv = 0;
		}
	}
	
	BY_HANDLE_FILE_INFORMATION FileInformation;
	if (rv == 0 && *lpBHandleIsUntrusted && GetFileInformationByHandle(handle, &FileInformation))
		if (FileInformation.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY || FileInformation.dwFileAttributes & FILE_ATTRIBUTE_DEVICE)
			*lpBHandleIsUntrusted = FALSE;
	return rv;
}



HANDLE untrustedUserToken = NULL;

HANDLE getUntrustedUserToken() {
	if (untrustedUserToken == NULL) {
		HANDLE primaryToken;
		if (LogonUser(PPI_UNTRUSTED_USER, L".", PPI_UNTRUSTED_USER_PASSWORD, LOGON32_LOGON_INTERACTIVE, LOGON32_PROVIDER_DEFAULT, &primaryToken)) {
			if (!DuplicateToken(primaryToken, SecurityIdentification, &untrustedUserToken))
				PPI_ERROR("Failed to duplicate token with impersonation requirement: error %d", GetLastError());
			CloseHandle(primaryToken);
		}
	}
	return untrustedUserToken;
}

GENERIC_MAPPING genericMapping;

PRIVILEGE_SET PrivilegeSet;


int pSDIsWritable(const PSECURITY_DESCRIPTOR pSD, HANDLE token, LPBOOL lpBwritable) {
	
	genericMapping.GenericRead = FILE_GENERIC_READ;
	genericMapping.GenericWrite = FILE_GENERIC_WRITE;
	genericMapping.GenericExecute = FILE_GENERIC_EXECUTE;
	genericMapping.GenericAll = FILE_ALL_ACCESS;

	
	DWORD dwPrivSetSize = sizeof(PRIVILEGE_SET);
	DWORD dwAccessDesired = FILE_WRITE_DATA;
	DWORD dwAccessAllowed;
	MapGenericMask(&dwAccessDesired, &genericMapping);
	
	if (0 == AccessCheck(pSD, token, dwAccessDesired, &genericMapping, &PrivilegeSet, &dwPrivSetSize, &dwAccessAllowed, lpBwritable)) {
		PPI_ERROR("AccessCheck failed with error %d", GetLastError());
		return -1;
	} else {
		return 0;
	}
}

int ppi_level_pSDIsUntrustedW(const PSECURITY_DESCRIPTOR pSD, LPCWSTR lpFileName, LPBOOL lpBpSDIsUntrusted) {

	int rv = -1;
	BOOL bNeedSlowCheck = TRUE;
	DWORD tempRes;

	
	if (pSDIsWritable(pSD, getUntrustedUserToken(), lpBpSDIsUntrusted) == 0) {
		bNeedSlowCheck = FALSE;
		rv = 0;
	}

	if (bNeedSlowCheck) {
slowCheck:
		if (ppi_level_pSDIsUntrustedAclCheck(pSD, lpBpSDIsUntrusted) == 0) {
			rv = 0;
			//AVOID PRINTING ANYTHING>>>>>>> VS may not open project!!!!
			//PPI_INFO("Checking with slow path for file %ls gave is untrusted: %d", lpFileName, *lpBpSDIsUntrusted);
		}
		else {
			//PPI_ERROR("Checking with slow path for file %ls failed", lpFileName);
		}
	}

	//Check if owned by untrusted user
	PSID pOwner;
	BOOL bOwnerDefaulted;

	//If it is determined to be benign, do the final check
	if (rv == 0 && !*lpBpSDIsUntrusted) {
		if (GetSecurityDescriptorOwner(pSD, &pOwner, &bOwnerDefaulted) == 0) {
			//PPI_ERROR("Failed to get owner info of file");
			rv = -1;
		} else {
			LPWSTR sidString;
			if (ConvertSidToStringSidW(pOwner, &sidString)) {
				if (wcscmp(sidString, PPI_UNTRUSTED_SIDSTRING) == 0)
					*lpBpSDIsUntrusted = TRUE;
				LocalFree(sidString);
			}
		}
	}

	if (lpFileName && (rv || *lpBpSDIsUntrusted)) {
		//if (ppi_utils_wstringInArray(lpFileName, benignToReadFiles)) {
		if (ppi_level_benign2read(lpFileName)) {
			rv = 0;
			*lpBpSDIsUntrusted = FALSE;
		}
	}
	//PPI_INFO("rv is %d, isUntrusted %d, file: %ls, pSD: %lp", rv, *lpBpSDIsUntrusted, lpFileName, pSD);
out:
	return rv;
}



int MULTIP ppi_level_pSDCanFlow2PrincipalW(const PSECURITY_DESCRIPTOR pSD, LPCWSTR lpFileName, LPBOOL lpBpSDIsUntrusted) {

	int rv = -1;
	BOOL bNeedSlowCheck = TRUE;
	DWORD tempRes;

	//FOR EACH UNTRUSTED USER TOKEN
	if (pSDIsWritable(pSD, getUntrustedUserToken(), lpBpSDIsUntrusted) == 0) {
		bNeedSlowCheck = FALSE;
		rv = 0;
	}

	if (bNeedSlowCheck) {
slowCheck:
		if (ppi_level_pSDIsUntrustedAclCheck(pSD, lpBpSDIsUntrusted) == 0)
			rv = 0;
	}
	
	if (lpFileName && (rv || *lpBpSDIsUntrusted)) {
		//if (ppi_utils_wstringInArray(lpFileName, benignToReadFiles)) {
		if (ppi_level_benign2read(lpFileName)) {
			rv = 0;
			*lpBpSDIsUntrusted = FALSE;
		}
	}
out:
	return rv;
}




int ppi_level_pSDIsUntrustedAclCheck(const PSECURITY_DESCRIPTOR pSD, LPBOOL lpbIsUntrusted) {

	int rv = -1;
	PACL pACL = NULL;
	ACL_SIZE_INFORMATION aclSizeInfo;
	PVOID pTempAce;
	BOOL bDaclPresent, bDaclDefaulted;
	unsigned int i;

	if (GetSecurityDescriptorDacl(pSD, &bDaclPresent, &pACL, &bDaclDefaulted) == 0)
		goto out;

	if (!bDaclPresent || pACL == NULL) {
		*lpbIsUntrusted = TRUE;
		rv = 0;
		goto out;
	}

	if (!GetAclInformation(pACL, (LPVOID)&aclSizeInfo, sizeof(ACL_SIZE_INFORMATION), AclSizeInformation))
		goto out;
	
	GENERIC_MAPPING genericMapping;
	genericMapping.GenericRead = FILE_GENERIC_READ;
	genericMapping.GenericWrite = FILE_GENERIC_WRITE;
	genericMapping.GenericExecute = FILE_GENERIC_EXECUTE;
	genericMapping.GenericAll = FILE_ALL_ACCESS;

	DWORD dwPrivSetSize = sizeof(PRIVILEGE_SET);
	DWORD dwAccessDesired = FILE_WRITE_DATA;
	MapGenericMask(&dwAccessDesired, &genericMapping);


	for (i=0; i< aclSizeInfo.AceCount; i++) {
		if (!GetAce(pACL, i, &pTempAce)) {
			goto out;
		}

		switch (((PACE_HEADER)pTempAce)->AceType){
		case ACCESS_ALLOWED_ACE_TYPE: {
			PACCESS_ALLOWED_ACE paaACE = (PACCESS_ALLOWED_ACE)pTempAce;
			LPTSTR sidString;
			if (ConvertSidToStringSid(&(paaACE->SidStart), &sidString)) {
				if (paaACE->Mask & dwAccessDesired) {
					if (ppi_utils_wstringInArray(sidString, untrustedInGroupSID)) {
						*lpbIsUntrusted = TRUE;
						rv = 0;
						LocalFree(sidString);
						goto out;
					}
				}
				LocalFree(sidString);
			}
			break;
										}
		default:
			PPI_ERROR("ACE type is %d, currently only handle ACCESS_ALLOWED_ACE_TYPE which is %d", ((PACE_HEADER)pTempAce)->AceType, ACCESS_ALLOWED_ACE_TYPE);
		}
	}

	*lpbIsUntrusted = FALSE;
	rv = 0;

out:
	return rv;
}



int ppi_level_OAExistsAndIsUntrustedW(OBJECT_ATTRIBUTES *ObjectAttributes, BOOL *bFileExists, BOOL *bFileIsUntrusted) {
	HANDLE handle;
	int rv;
	IO_STATUS_BLOCK IoStatusBlock;
	rv = OriginalNtCreateFile(&handle, READ_CONTROL, ObjectAttributes, &IoStatusBlock, NULL,
			FILE_ATTRIBUTE_NORMAL, FILE_SHARE_READ, FILE_OPEN, 0, NULL, 0);

	if (rv == STATUS_SUCCESS) {
		*bFileExists = TRUE;
		//OPT: pass in file name???
		rv = ppi_getHandleIsUntrustedW(handle, NULL, bFileIsUntrusted);
		//CloseHandle(handle);
		OriginalNtClose(handle);
		return 0;
	} else if (rv == STATUS_OBJECT_NAME_NOT_FOUND) {
		*bFileExists = FALSE;
		return 0;
	} else {
		PPI_CRITICAL("Unknown status on OA: %Xh", rv);
		return -1;
	}
}



int ppi_level_fileExistsAndIsUntrustedW(LPCWSTR lpFileName, BOOL *bFileExists, BOOL *bFileIsUntrusted) {
	int rv = -1;

	if (ppi_utils_fileExist(lpFileName, bFileExists)) {
		PPI_ERROR("Failed to check if file to be copied exists or not. %ls", lpFileName);
		goto out;
	}
	if (*bFileExists) {
		if (ppi_getFileIsUntrustedW(lpFileName, bFileIsUntrusted)) {
			PPI_ERROR("Failed to check if existing file %ls is untrusted", lpFileName);
			goto out;
		}
	}
	rv = 0;
out:
	return rv;
}

int ppi_level_fileExistsAndIsUntrustedA(LPCSTR lpFileName, BOOL *bFileExists, BOOL *bFileIsUntrusted) {
	int rv;

	wchar_t *buffer = ppi_bm_path_malloc();
	if (buffer) {
		if (ppi_utils_convertChar2Wchar(lpFileName, buffer, PPI_PATH_WCHAR_SIZE)) {
			rv = -1;
			goto out;
		}
		rv = ppi_level_fileExistsAndIsUntrustedW(buffer, bFileExists, bFileIsUntrusted);
	}
out:
	if (buffer)
		ppi_bm_free(buffer);
	return rv;
}





BOOL ppi_level_downgradeHandle(HANDLE handle) {
	
	BOOL result = FALSE;
	PACL pACL = NULL, pNewAcl = NULL;
	PSECURITY_DESCRIPTOR pSD = NULL, psdNew = NULL;
	ACL_SIZE_INFORMATION aclSizeInfo;
	PEXPLICIT_ACCESS pListOfExplicitEntries = NULL;
	PVOID pTempAce;

	DWORD dwNewAclSize;
	PSID psid = NULL;

	unsigned int i;

	
	DWORD dwRes = GetSecurityInfo(handle, 
		SE_FILE_OBJECT, DACL_SECURITY_INFORMATION|OWNER_SECURITY_INFORMATION|GROUP_SECURITY_INFORMATION, NULL, NULL, &pACL, NULL, &pSD);
		
	ULONG countOfExplicitEntries = 0;

	if (ERROR_SUCCESS != dwRes) {
		PPI_ERROR("Failed at GetNamedSecurityInfo for handle %x, errno %u %u\n", handle, dwRes, GetLastError());
	} else {
		if (pACL == NULL)
			goto cleanup;


		if (!GetAclInformation(pACL, (LPVOID)&aclSizeInfo, sizeof(ACL_SIZE_INFORMATION), AclSizeInformation))
			goto cleanup;

		if (!ConvertStringSidToSid(PPI_UNTRUSTED_SIDSTRING, &psid)) {
			PPI_ERROR("Failed to convert string to sid\n");
			goto cleanup;
		}
		
		dwNewAclSize = aclSizeInfo.AclBytesInUse + sizeof(ACCESS_ALLOWED_ACE) + GetLengthSid(psid) - sizeof(DWORD);

		pNewAcl = (PACL)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, dwNewAclSize);
		if (pNewAcl == NULL)
			goto cleanup;

		if (!InitializeAcl(pNewAcl, dwNewAclSize, ACL_REVISION))
			goto cleanup;

		if (aclSizeInfo.AceCount) {
			for (i=0; i< aclSizeInfo.AceCount; i++) {
				if (!GetAce(pACL, i, &pTempAce))
					goto cleanup;
				if (!AddAce(pNewAcl, ACL_REVISION, MAXDWORD, pTempAce, ((PACE_HEADER)pTempAce)->AceSize))
					goto cleanup;
			}
		}
		if (!AddAccessAllowedAce(pNewAcl, ACL_REVISION, FILE_ALL_ACCESS, psid)) {
			PPI_ERROR("Failed at AddAccessAllowedAce\n");
			goto cleanup;
		}

		int rv = SetSecurityInfo(handle, SE_FILE_OBJECT, DACL_SECURITY_INFORMATION, NULL, NULL, pNewAcl, NULL);
		if (ERROR_SUCCESS != rv) {
			PPI_ERROR("Failed to set new acl: rv is %d\n", rv);
			goto cleanup;
		}

		result = TRUE;
	}
cleanup:
	if (pSD != NULL)
		LocalFree((HLOCAL)pSD);
	if (psdNew != NULL)
		LocalFree((HLOCAL)psdNew);
	if (psid != NULL)
		LocalFree((HLOCAL)psid);
	if (pNewAcl != NULL)
		LocalFree((HLOCAL)pNewAcl);

	return result;
}

BOOL ppi_level_downgradeFile(LPWSTR filename) {
	//TODO: Add ace to untrusted files created
	BOOL result = FALSE;
	PACL pACL = NULL, pNewAcl = NULL;
	PSECURITY_DESCRIPTOR pSD = NULL, psdNew = NULL;
	ACL_SIZE_INFORMATION aclSizeInfo;
	PEXPLICIT_ACCESS pListOfExplicitEntries = NULL;
	PVOID pTempAce;

	DWORD dwNewAclSize;
	PSID psid = NULL;

	unsigned int i;
	

	DWORD dwRes = GetNamedSecurityInfo(filename, 
		SE_FILE_OBJECT, DACL_SECURITY_INFORMATION|OWNER_SECURITY_INFORMATION|GROUP_SECURITY_INFORMATION, NULL, NULL, &pACL, NULL, &pSD);
		
	ULONG countOfExplicitEntries = 0;

	if (ERROR_SUCCESS != dwRes) {
		PPI_ERROR("Failed at GetNamedSecurityInfo for file %ls, errno %u %u\n", filename, dwRes, GetLastError());
	} else {
		if (pACL == NULL)
			goto cleanup;

		if (!GetAclInformation(pACL, (LPVOID)&aclSizeInfo, sizeof(ACL_SIZE_INFORMATION), AclSizeInformation))
			goto cleanup;

		if (!ConvertStringSidToSid(PPI_UNTRUSTED_SIDSTRING, &psid)) {
			PPI_ERROR("Failed to convert string to sid\n");
			goto cleanup;
		}
		
		dwNewAclSize = aclSizeInfo.AclBytesInUse + sizeof(ACCESS_ALLOWED_ACE) + GetLengthSid(psid) - sizeof(DWORD);

		pNewAcl = (PACL)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, dwNewAclSize);
		if (pNewAcl == NULL)
			goto cleanup;

		if (!InitializeAcl(pNewAcl, dwNewAclSize, ACL_REVISION))
			goto cleanup;

		if (aclSizeInfo.AceCount) {
			for (i=0; i< aclSizeInfo.AceCount; i++) {
				if (!GetAce(pACL, i, &pTempAce))
					goto cleanup;
				if (!AddAce(pNewAcl, ACL_REVISION, MAXDWORD, pTempAce, ((PACE_HEADER)pTempAce)->AceSize))
					goto cleanup;
			}
		}
		if (!AddAccessAllowedAce(pNewAcl, ACL_REVISION, FILE_ALL_ACCESS, psid)) {
			PPI_ERROR("Failed at AddAccessAllowedAce\n");
			goto cleanup;
		}

		if (ERROR_SUCCESS != SetNamedSecurityInfo(filename, SE_FILE_OBJECT, DACL_SECURITY_INFORMATION, NULL, NULL, pNewAcl, NULL)) {
			PPI_ERROR("Failed to set new acl\n");
			goto cleanup;
		}

		result = TRUE;
	}
cleanup:
	if (pSD != NULL)
		LocalFree((HLOCAL)pSD);
	if (psdNew != NULL)
		LocalFree((HLOCAL)psdNew);
	if (psid != NULL)
		LocalFree((HLOCAL)psid);
	if (pNewAcl != NULL)
		LocalFree((HLOCAL)pNewAcl);

	return result;
}



BOOL ppi_level_downgradeDirectory(LPWSTR filename) {
	//TODO: Add ace to untrusted files created
	BOOL result = FALSE;
	PACL pACL = NULL, pNewAcl = NULL;
	PSECURITY_DESCRIPTOR pSD = NULL, psdNew = NULL;
	ACL_SIZE_INFORMATION aclSizeInfo;
	PEXPLICIT_ACCESS pListOfExplicitEntries = NULL;
	PVOID pTempAce;

	DWORD dwNewAclSize;
	PSID psid = NULL;

	unsigned int i;
	

	DWORD dwRes = GetNamedSecurityInfo(filename, 
		SE_FILE_OBJECT, DACL_SECURITY_INFORMATION|OWNER_SECURITY_INFORMATION|GROUP_SECURITY_INFORMATION, NULL, NULL, &pACL, NULL, &pSD);
		
	ULONG countOfExplicitEntries = 0;

	if (ERROR_SUCCESS != dwRes) {
		PPI_ERROR("Failed at GetNamedSecurityInfo, errno %u %u\n", dwRes, GetLastError());
	} else {
		if (pACL == NULL)
			goto cleanup;

		if (!GetAclInformation(pACL, (LPVOID)&aclSizeInfo, sizeof(ACL_SIZE_INFORMATION), AclSizeInformation))
			goto cleanup;

		if (!ConvertStringSidToSid(PPI_UNTRUSTED_SIDSTRING, &psid)) {
			PPI_ERROR("Failed to convert string to sid\n");
			goto cleanup;
		}
		
		dwNewAclSize = aclSizeInfo.AclBytesInUse + sizeof(ACCESS_ALLOWED_ACE) + GetLengthSid(psid) - sizeof(DWORD);

		pNewAcl = (PACL)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, dwNewAclSize);
		if (pNewAcl == NULL)
			goto cleanup;

		if (!InitializeAcl(pNewAcl, dwNewAclSize, ACL_REVISION))
			goto cleanup;

		if (aclSizeInfo.AceCount) {
			for (i=0; i< aclSizeInfo.AceCount; i++) {
				if (!GetAce(pACL, i, &pTempAce))
					goto cleanup;
				if (!AddAce(pNewAcl, ACL_REVISION, MAXDWORD, pTempAce, ((PACE_HEADER)pTempAce)->AceSize))
					goto cleanup;
			}
		}
		if (!AddAccessAllowedAceEx(pNewAcl, ACL_REVISION, CONTAINER_INHERIT_ACE|OBJECT_INHERIT_ACE, FILE_ALL_ACCESS|FILE_LIST_DIRECTORY|FILE_ADD_FILE, psid)) {
			PPI_ERROR("Failed at AddAccessAllowedAce\n");
			goto cleanup;
		}

		if (ERROR_SUCCESS != SetNamedSecurityInfo(filename, SE_FILE_OBJECT, DACL_SECURITY_INFORMATION, NULL, NULL, pNewAcl, NULL)) {
			PPI_ERROR("Failed to set new acl\n");
			goto cleanup;
		}

		result = true;
	}
cleanup:
	if (pSD != NULL)
		LocalFree((HLOCAL)pSD);
	if (psdNew != NULL)
		LocalFree((HLOCAL)psdNew);
	if (psid != NULL)
		LocalFree((HLOCAL)psid);
	
	if (pNewAcl != NULL)
		LocalFree((HLOCAL)pNewAcl);

	return result;
}



PPI_Resource::PPI_Resource(HANDLE h, PPI_OA *ppi_oa): _handle(h), _ppi_oa(ppi_oa){
	_pSD = NULL;
	_ppi_bm_managePSD = true;
	path_ = NULL;
};

PPI_Resource::PPI_Resource(const wchar_t *path): _handle(NULL) {
	_ppi_oa = NULL;
	path_ = ppi_bm_path_malloc();
	_pSD = NULL;
	wcsncpy(path_, path, PPI_PATH_WCHAR_SIZE);
};

// With respect to a principal, tells if a resource is untrusted.
// If the resource is untrusted, returns which prinicpal not trusted by the given principal can write into the resource.
// Otherwise, returns the original principal.
const PPI_PRINCIPAL *PPI_Resource::getUntrustedOwner(const PPI_PRINCIPAL *p) {
	
	const PPI_PRINCIPAL *result = p;
	const list<const PPI_PRINCIPAL *> *untrustedPrincipals = p->getUntrustedPrincipals();

	BOOL writableByUntrustedP = FALSE;
	PSECURITY_DESCRIPTOR pSD = _getPSD();
	const wchar_t *path = (_ppi_oa ? _ppi_oa->getPrintablePath() : path_);
	bool isTrusted = true;


	if (pSD == NULL) {
		PPI_CRITICAL("cannot get psd for the resource %ls", path);
		return p;
	}
	for (list<const PPI_PRINCIPAL *>::const_iterator it = untrustedPrincipals->begin(); it != untrustedPrincipals->end(); it++) {
		if (pSDIsWritable(pSD, (*it)->getToken(), &writableByUntrustedP) == 0) {
			if (writableByUntrustedP) {
				isTrusted = false;
				result = (*it);
				break;
			}
		} else {
		}

	}

	//TODO: Check ownership
	if (isTrusted && path && !ppi_principal_is_Untrusted(p)) {
		BOOL handleIsUntrusted;
		if (ppi_level_getMSFileIsUntrustedW(path, &handleIsUntrusted)) {
			PPI_ERROR("Failed to get whether file %ls is untrusted from MS integrity model. Assuming trusted", path);
		} else if (handleIsUntrusted && !p->acceptMSLowIntegrityFile()) {
			isTrusted = false;
			result = PPI_PRINCIPAL_UNTRUSTED_PRINCIPAL;
			PPI_HIGH_VIOLATION("Resource is of MS Low integrity");
		}
	}

	if (path) {
		const PPI_PRINCIPAL *overridedP = PPI_GET_OVERRIDED_OWNERSHIP(path);
		if (overridedP != nullptr) {
			for (list<const PPI_PRINCIPAL *>::const_iterator it = p->getTrustedPrincipals()->begin(); it != p->getTrustedPrincipals()->end(); it++)
				if (overridedP == *it)
					return p;
			result = overridedP;
		}
		
	}
	
	return result;
}


bool PPI_Resource::isTrustedByPrincipal(const PPI_PRINCIPAL *p) {

	// If none of the principals not trusted by p can write into the resource,
	// then it is trusted.
	return p == getUntrustedOwner(p);
}


PSECURITY_DESCRIPTOR PPI_Resource::_getPSD() {

	PSECURITY_DESCRIPTOR pSD = NULL;

	if (_pSD == NULL) {
		
		BOOL tempRes;
		DWORD lenNeeded = PPI_BUFFERPOOL_PER_ENTRY_SIZE;
	
		const wchar_t *lpFileName = NULL;
		if (_ppi_oa != NULL)
			lpFileName = _ppi_oa->getPrintablePath();
		else
			lpFileName = path_;

		if (_handle) {
			pSD = (PSECURITY_DESCRIPTOR)ppi_bm_malloc(lenNeeded);
			tempRes = GetKernelObjectSecurity(_handle, DACL_SECURITY_INFORMATION|GROUP_SECURITY_INFORMATION|OWNER_SECURITY_INFORMATION, NULL, 0, &lenNeeded);
			if (tempRes) {
				_pSD = pSD;
				goto out;
			}
			ppi_bm_free(pSD);
			if(GetLastError() == ERROR_INSUFFICIENT_BUFFER) {
				pSD = (PSECURITY_DESCRIPTOR)ppi_bm_malloc(lenNeeded);
				if (pSD != NULL) {
					tempRes = GetKernelObjectSecurity(_handle, DACL_SECURITY_INFORMATION|GROUP_SECURITY_INFORMATION|OWNER_SECURITY_INFORMATION, pSD, lenNeeded, &lenNeeded);
					if (tempRes) {
						_pSD = pSD;
						goto out;
					} else if (GetLastError() != ERROR_INVALID_FUNCTION) {
						PPI_ERROR("Failed to perform GetKernelObjectSecurity on file %ls handle: %lu due to unknown reason: %d", lpFileName, _handle, GetLastError());
					}
					ppi_bm_free(pSD);
				}
			}
		}
		/* Get information without using handle */
		if (lpFileName == NULL)	goto out;
		
		pSD = NULL;
		DWORD dwRes = GetNamedSecurityInfo(lpFileName, 
			SE_FILE_OBJECT, DACL_SECURITY_INFORMATION|OWNER_SECURITY_INFORMATION|GROUP_SECURITY_INFORMATION, NULL, NULL, NULL, NULL, &pSD);

		if (ERROR_SUCCESS != dwRes) {
			PPI_ERROR("Benign process failed to call GetKernelObjectSecurity and GetNamedSecurityInfo on file %ls due to permission reason... Assuming file is benign", lpFileName);
		} else {
			_pSD = pSD;
			_ppi_bm_managePSD = false;
			goto out;
		}
		
	}
out:
	return _pSD;
}
