//
// Portable Integrity Protection (PIP) System -
// Copyright (C) 2015 Secure Systems Laboratory, Stony Brook University
//
// This file is part of Portable Integrity Protection (PIP) System.
//
// Portable Integrity Protection (PIP) System is free software: you can
// redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Portable Integrity Protection (PIP) System is distributed in the hope that it
// will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Portable Integrity Protection (PIP) System.  If not, see
// <http://www.gnu.org/licenses/>.
//

#include "ppi_packet.h"
#include "ppi_helperConnection.h"
#include "ppi_principal.h"
#include "ppi_bufferManager.h"

#include "ppi_debug.h"
#include <Sddl.h>

#include <algorithm>
#include <set>
#include <string>

/* Manage principal */

//List of programs trusted to consume MS Low integrity data
const std::wstring ppi_trustedToReadMSLowIntegrity[] = {
	L"C:\\Program Files\\Internet Explorer\\iexplore.exe"
};

const std::set<std::wstring> ppi_trustedToReadMSLowIntegritySet(std::begin(ppi_trustedToReadMSLowIntegrity), std::end(ppi_trustedToReadMSLowIntegrity));

PPI_PRINCIPAL::PPI_PRINCIPAL(wchar_t *fullsid, wchar_t *sid, wchar_t *username, wchar_t *password, wchar_t *redirect_prefix, int id, const wchar_t *programPath): fullsid(fullsid), sid(sid), username(username), password(password), redirect_prefix(redirect_prefix), id(id) {
	_token = NULL;
	if (ppi_trustedToReadMSLowIntegritySet.find(wstring(programPath)) != ppi_trustedToReadMSLowIntegritySet.end())
		canAcceptMSLowIntegrity = true;
	else
		canAcceptMSLowIntegrity = false;
};

const PPI_PRINCIPAL *PPI_PRINCIPAL::getLessTrusted(const PPI_PRINCIPAL *p1, const PPI_PRINCIPAL *p2) const{
	if (cachedUntrustedOrder.find(p1) == cachedUntrustedOrder.end())
		return p2;	
	if (cachedUntrustedOrder.find(p2) == cachedUntrustedOrder.end())
		return p1;
	if (cachedUntrustedOrder.at(p1) < cachedUntrustedOrder.at(p2))
		return p1;
	return p2;
}

void PPI_PRINCIPAL::addUntrustedPrincipal(const PPI_PRINCIPAL *p) {
	this->untrustedPrincipals.push_back(p);
	cachedUntrustedOrder[p] = cachedUntrustedOrder.size();
}

void PPI_PRINCIPAL::addTrustedPrincipal(const PPI_PRINCIPAL *p) {
	this->trustedPrincipals.push_back(p);
}

bool PPI_PRINCIPAL::acceptInformationFrom(const PPI_PRINCIPAL *p) const {
	return (find(this->untrustedPrincipals.begin(), this->untrustedPrincipals.end(), p) != this->untrustedPrincipals.end());
}

bool PPI_PRINCIPAL::acceptMSLowIntegrityFile() const {
	return canAcceptMSLowIntegrity;
}
	
	
HANDLE PPI_PRINCIPAL::getToken() const {
	if (_token == NULL) {

		//Experiment: Obtaining tokens from helper
		/*
		ppi_pkt_prepare_packets(DupToken, pkt, response);
		pkt.id = this->id;
			
		if (ppi_helper_sendRecvCorrectly(pkt, response)) {
			if (response.l_rv == STATUS_SUCCESS) {
				_token = response.token;
				goto out;
			}
		}
		PPI_CRITICAL("Failed to duplicate token from helper");
		*/

		HANDLE handle = NULL;
		HANDLE primaryToken;
		if (LogonUser(username, L".", password, LOGON32_LOGON_INTERACTIVE, LOGON32_PROVIDER_DEFAULT, &primaryToken)) {
			if (!DuplicateToken(primaryToken, SecurityIdentification, &handle))
				PPI_CRITICAL("Failed to duplicate token with impersonation requirement: error %d", GetLastError());
			else
				_token = handle;
			CloseHandle(primaryToken);
		}
		
	}
out:
	return _token;
}


const list<const PPI_PRINCIPAL *> *PPI_PRINCIPAL::getUntrustedPrincipals() const {
	return &untrustedPrincipals;
}

const list<const PPI_PRINCIPAL *> *PPI_PRINCIPAL::getTrustedPrincipals() const {
	return &trustedPrincipals;
}


#define PRINCIPAL_INFO_LIST(x) \
	PPI_MULTIP_ ##x ##_SIDSTRING, \
	PPI_MULTIP_ ##x ##_ID_STR, \
	PPI_MULTIP_ ##x ##_USER, \
	PPI_MULTIP_ ##x ##_PASSWORD, \
	PPI_MULTIP_ ##x ##_REDIRECT_PREFIX, \
	PPI_MULTIP_ ##x ##_ID


void PPI_PRINCIPAL_MANAGER::initialize() {

	allPrincipals.insert(new PPI_PRINCIPAL(PRINCIPAL_INFO_LIST(BENIGN)));
	allPrincipals.insert(new PPI_PRINCIPAL(PRINCIPAL_INFO_LIST(UNTRUSTED)));

	PPI_PRINCIPAL *benignP = _getPrincipalFromID(PPI_MULTIP_BENIGN_ID);
	PPI_PRINCIPAL *untrustedP = _getPrincipalFromID(PPI_MULTIP_UNTRUSTED_ID);
	
	if (benignP == NULL || untrustedP == NULL)
		return;
	
#ifndef PPI_TWO_PRINCIPAL
	allPrincipals.insert(new PPI_PRINCIPAL(PRINCIPAL_INFO_LIST(MICROSOFT)));
	PPI_PRINCIPAL *microsoftP = _getPrincipalFromID(PPI_MULTIP_MICROSOFT_ID);

	if (microsoftP == NULL)
		return;
#endif

#ifdef PPI_TWO_PRINCIPAL

	benignP->addUntrustedPrincipal(untrustedP);
	benignP->addTrustedPrincipal(benignP);

	untrustedP->addTrustedPrincipal(untrustedP);
	untrustedP->addTrustedPrincipal(benignP);

#else

	benignP->addUntrustedPrincipal(untrustedP);
	benignP->addUntrustedPrincipal(microsoftP);

	microsoftP->addUntrustedPrincipal(untrustedP);
	
	benignP->addTrustedPrincipal(benignP);

	microsoftP->addTrustedPrincipal(microsoftP);
	microsoftP->addTrustedPrincipal(benignP);
	
	untrustedP->addTrustedPrincipal(untrustedP);
	untrustedP->addTrustedPrincipal(microsoftP);
	untrustedP->addTrustedPrincipal(benignP);
#endif
}

const PPI_PRINCIPAL *PPI_PRINCIPAL_MANAGER::getPrincipalFromID(int id) const {
	for (unordered_set<PPI_PRINCIPAL *>::iterator it = allPrincipals.begin(); it != allPrincipals.end(); ++it) {
		if (id == (*it)->id)
			return *it;
	}
	return getPrincipalFromID(PPI_MULTIP_BENIGN_ID);
}

PPI_PRINCIPAL *PPI_PRINCIPAL_MANAGER::_getPrincipalFromID(int id) {
	for (unordered_set<PPI_PRINCIPAL *>::iterator it = allPrincipals.begin(); it != allPrincipals.end(); ++it) {
		if (id == (*it)->id)
			return *it;
	}
	PPI_CRITICAL("Retrieving Principal using non-existent ID: %d. Assuming to be benign", id);
	return _getPrincipalFromID(PPI_MULTIP_BENIGN_ID);
}


const PPI_PRINCIPAL *PPI_PRINCIPAL_MANAGER::getCurrentPrincipal(){
	if (current_principal == NULL) {

		HANDLE hProcess = GetCurrentProcess();
		HANDLE hProcessToken = NULL;
		int rv = -1;
		PTOKEN_USER tokenInformation = NULL;
		DWORD dwProcessTokenInfoAllocSize;
		
		wchar_t *sid = (wchar_t*) ppi_bm_default_malloc();
		if (sid == NULL) {
			PPI_CRITICAL("Buffer is Null");
			goto out;
		}

		if (!OpenProcessToken(hProcess, TOKEN_READ, &hProcessToken) || !hProcessToken ) {
			PPI_ERROR("Failed to open process token: %d", GetLastError());
			goto out;
		}

		tokenInformation = (PTOKEN_USER) ppi_bm_default_malloc();
		if (tokenInformation != NULL) {
			if (GetTokenInformation(hProcessToken, TokenUser, tokenInformation, ppi_bm_default_mallocSizeInByte, &dwProcessTokenInfoAllocSize)) {
				LPWSTR sidString;
				if (ConvertSidToStringSid(tokenInformation->User.Sid, &sidString)) {
					for (unordered_set<PPI_PRINCIPAL *>::iterator it = allPrincipals.begin(); it != allPrincipals.end(); ++it) {
						//PPI_CRITICAL("Comparing sid: %ls with it: %ls", sidString, (*it)->fullsid);
						if (wcscmp(sidString, (*it)->fullsid) == 0) {
							current_principal = (*it);
							break;
						}
					}
					rv = 0; 
					LocalFree(sidString);
				}
			} else {
				if (GetLastError() == ERROR_INSUFFICIENT_BUFFER) {
					PPI_CRITICAL("Default malloc size is not enough");
				} else {
					PPI_CRITICAL("Failed to retreive token information due to unknown error: %x", GetLastError());
				}
			}
			ppi_bm_free(tokenInformation);
		}
		ppi_callOriginal(NtClose, hProcessToken);
		ppi_bm_free(sid);

		if (current_principal == NULL) {
			current_principal = _getPrincipalFromID(PPI_MULTIP_BENIGN_ID);
			PPI_CRITICAL("Assuming current principal is BENIGN");
		}
	}
out:
	return current_principal;
}
