/* * Portable Integrity Protection (PIP) System -
 * Copyright (C) 2015 Secure Systems Laboratory, Stony Brook University
 *
 * This file is part of Portable Integrity Protection (PIP) System.
 *
 * Portable Integrity Protection (PIP) System is free software: you can
 * redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Portable Integrity Protection (PIP) System is distributed in the hope that it
 * will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Portable Integrity Protection (PIP) System.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#ifndef __PPI_PRINCIPAL_H__
#define __PPI_PRINCIPAL_H__


#include <Windows.h>
#include <list>
#include <unordered_set>
#include <unordered_map>
#include <set>
#include "ppi_utils.h"


/* Manage principal */


using namespace std;

class PPI_PRINCIPAL {
public:
	const wchar_t *fullsid;
	const wchar_t *sid;
	const wchar_t *username;
	const wchar_t *password;
	const wchar_t *redirect_prefix;
	const int id;
	const list<const PPI_PRINCIPAL *> *getUntrustedPrincipals() const;
	const list<const PPI_PRINCIPAL *> *getTrustedPrincipals() const;

	bool acceptInformationFrom(const PPI_PRINCIPAL *p) const;
	bool canWriteTo(HANDLE h);

	PPI_PRINCIPAL(wchar_t *fullsid, wchar_t *sid, wchar_t *username, wchar_t *password, wchar_t *redirect_prefix, int id, const wchar_t *programPath=ppi_utils_getProgramPath());

	HANDLE getToken() const;

	~PPI_PRINCIPAL() {
		if (_token != NULL)
			CloseHandle(_token);
	}

	void addUntrustedPrincipal(const PPI_PRINCIPAL *p);
	void addTrustedPrincipal(const PPI_PRINCIPAL *p);
	const PPI_PRINCIPAL *getLessTrusted(const PPI_PRINCIPAL *p1, const PPI_PRINCIPAL *p2) const;
	bool PPI_PRINCIPAL::acceptMSLowIntegrityFile() const;


private:
	list<const PPI_PRINCIPAL *> untrustedPrincipals;
	list<const PPI_PRINCIPAL *> trustedPrincipals; //For maintaining order of checking
	unordered_map<const PPI_PRINCIPAL *, int> cachedUntrustedOrder;
	mutable HANDLE _token;
	bool canAcceptMSLowIntegrity;
};


class PPI_PRINCIPAL_MANAGER {
public:
	static PPI_PRINCIPAL_MANAGER &getInstance() {
		static PPI_PRINCIPAL_MANAGER instance;
		static bool initialized = false;
		if (!initialized) {
			instance.initialize();
			initialized = true;
		}
		return instance;
	}
	const PPI_PRINCIPAL *getCurrentPrincipal();
	const PPI_PRINCIPAL *getPrincipalFromID(int id) const;

	~PPI_PRINCIPAL_MANAGER() {
		for (unordered_set<PPI_PRINCIPAL *>::iterator it = allPrincipals.begin(); it != allPrincipals.end(); it++)
			delete *it;
	}



private:
	PPI_PRINCIPAL_MANAGER() {};
	PPI_PRINCIPAL_MANAGER(PPI_PRINCIPAL_MANAGER const &);
	void operator=(PPI_PRINCIPAL_MANAGER const &);
	PPI_PRINCIPAL *current_principal;
	void initialize();
	unordered_set<PPI_PRINCIPAL *> allPrincipals;
	PPI_PRINCIPAL *_getPrincipalFromID(int id);
	

};



#define ppi_principal_isUntrusted() (PPI_PRINCIPAL_MANAGER::getInstance().getCurrentPrincipal()->id == PPI_MULTIP_UNTRUSTED_ID)

#define PPI_PRINCIPAL_IS_BENIGN() (PPI_PRINCIPAL_MANAGER::getInstance().getCurrentPrincipal()->id == PPI_MULTIP_BENIGN_ID)
#define ppi_principal_is_Untrusted(_p) (_p->id == PPI_MULTIP_UNTRUSTED_ID)

#define PPI_PRINCIPAL_UNTRUSTED_PRINCIPAL (PPI_PRINCIPAL_MANAGER::getInstance().getPrincipalFromID(PPI_MULTIP_UNTRUSTED_ID))

#define PPI_PRINCIPAL_GET_PRINCIPAL(_x) (PPI_PRINCIPAL_MANAGER::getInstance().getPrincipalFromID(_x ##_ID))

#define PPI_PRINCIPAL_GET_CURRENT PPI_PRINCIPAL_MANAGER::getInstance().getCurrentPrincipal()


#endif /* __PPI_PRINCIPAL_H__ */
