//
// Portable Integrity Protection (PIP) System -
// Copyright (C) 2015 Secure Systems Laboratory, Stony Brook University
//
// This file is part of Portable Integrity Protection (PIP) System.
//
// Portable Integrity Protection (PIP) System is free software: you can
// redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Portable Integrity Protection (PIP) System is distributed in the hope that it
// will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Portable Integrity Protection (PIP) System.  If not, see
// <http://www.gnu.org/licenses/>.
//

#include "ppi_config.h"
#include "ppi_debug.h"
#include "ppi_calls.h"
#include "ppi_utils.h"
#include "ppi_redirect.h"
#include "ppi_bufferManager.h"
#include "ppi_principal.h"
#include "ppi_regHandleManager.h"

/* Manage shadowing */

bool ppi_isFullPathW(const wchar_t *path);

const wchar_t *ppi_redirectablePath[] = {
	/*L"C:\\Documents and Settings\\Ubuntu\\Application Data\\",
	L"C:\\Documents and Settings\\Ubuntu\\Local Settings\\",
	L"C:\\Documents and Settings\\Ubuntu\\LOCALS~1\\",
	L"C:\\Documents and Settings\\Ubuntu\\Cookies\\",
	//L"C:\\Documents and Settings\\Ubuntu\\Desktop\\",
	L"C:\\DOCUME~1\\Ubuntu\\APPLIC~1\\",

	//Win 8.1*/
	L"C:\\Users\\" PPI_MULTIP_BENIGN_USER L"\\AppData\\",
	NULL};



bool ppi_isFullPathW(const wchar_t *path) {

	int colIdx = 0;

	if (path == NULL) {
		PPI_ERROR("shadowPath is NULL");
		return false;
	}

	while (path[colIdx] != L'\0' && path[colIdx] != L':' && colIdx < 10)
		colIdx++;

	if (path[colIdx] != L':') {
		PPI_ERROR("Requesting to convert relative path %ls to shadown path. Currently not supported", path);
		return false;
	}
	
	return true;
}


bool ppi_recursivelyCreateDirectoriesIgnoreLastW(const wchar_t *fullPath) {
	
	wchar_t *buf = ppi_bm_path_malloc();
	int i=0;
	
	if (buf != NULL) {
		if (wcsncmp(fullPath, L"\\??\\", 4) == 0) {
			fullPath += 4;
		}
	
		while (fullPath[i] != NULL) {
			if (fullPath[i] == L'\\' || fullPath[i] == L'/') {
				buf[i] = NULL;
				if ((CreateDirectory(buf, NULL) == 0) && (GetLastError() != ERROR_ALREADY_EXISTS)) {
					PPI_ERROR("Failed to create directory %ls for unknown reason: %d", buf, GetLastError());
					ppi_bm_free(buf);
					return false;
				}
			}
			buf[i] = fullPath[i];
			i++;
		}
		ppi_bm_free(buf);
	}

	return true;

}


void ppi_deallocShadowPath(wchar_t *shadowPath) {
	if (shadowPath != NULL)
		ppi_bm_free(shadowPath);
}


UNICODE_STRING *ppi_convertUnicode2BenignRegWAlloc(const UNICODE_STRING *originalPath) {
	
	wchar_t *fileName = originalPath->Buffer;
	size_t extraPrefixBytes = 0;

	if (fileName == NULL)
		return NULL;
		
	const wchar_t *firstOccurance = ppi_firstUntrustedSID(fileName, originalPath->Length/sizeof(wchar_t));
		//ppi_wcsnstr(fileName, PPI_UNTRUSTED_SIDSTRING, originalPath->Length/sizeof(wchar_t));

	if (firstOccurance == NULL) {
		return NULL;
	}
	
	wchar_t *shadowPath = ppi_bm_path_malloc();
	if (shadowPath == NULL) return NULL;

	UNICODE_STRING *shadowPathUnicode = (UNICODE_STRING *)ppi_bm_malloc(sizeof(UNICODE_STRING));
	if (shadowPathUnicode == NULL) {
		ppi_bm_free(shadowPath);
		return NULL;
	}

	shadowPathUnicode->Buffer = shadowPath;

	int offsetB = (firstOccurance-fileName)*sizeof(wchar_t);

	wcsncpy_s(shadowPath, PPI_PATH_WCHAR_SIZE, originalPath->Buffer, (originalPath->Length - offsetB)/sizeof(wchar_t));
	wcsncpy_s(shadowPath + offsetB/sizeof(wchar_t), PPI_PATH_WCHAR_SIZE - offsetB/sizeof(wchar_t), PPI_BENIGN_SIDSTRING, wcslen(PPI_BENIGN_SIDSTRING));
	wcsncpy_s(shadowPath + offsetB/sizeof(wchar_t) + wcslen(PPI_BENIGN_SIDSTRING), PPI_PATH_WCHAR_SIZE - offsetB/sizeof(wchar_t) - wcslen(PPI_BENIGN_SIDSTRING), fileName + offsetB/sizeof(wchar_t) + wcslen(PPI_BENIGN_SIDSTRING), (originalPath->Length - offsetB)/sizeof(wchar_t) - wcslen(PPI_BENIGN_SIDSTRING));
	shadowPath[originalPath->Length/sizeof(wchar_t)+1] = L'\0';
	
	shadowPathUnicode->Length = shadowPathUnicode->MaximumLength = originalPath->Length;
	
	return shadowPathUnicode;
}


int ppi_insertRedirectPrefix(const wchar_t *originalPath, const wchar_t *redirectPrefix, wchar_t *buffer, size_t bufferSizeInW) {
	
	wchar_t *curShadowPath = buffer;
	const wchar_t *origPtr = originalPath;

	if (originalPath == NULL)
		return -1;

	if (redirectPrefix == NULL || wcslen(redirectPrefix) == 0) {
		wcsncpy_s(buffer, bufferSizeInW, originalPath, wcslen(originalPath));
		buffer[wcslen(originalPath)] = L'\0';
		return 0;
	}
	
	BOOL firstColon = FALSE;
	while (((curShadowPath - buffer) < bufferSizeInW) && (*origPtr != NULL)) {
		*curShadowPath = *origPtr;
		curShadowPath++;
		if (!firstColon && *origPtr == L':') {
			curShadowPath  += _snwprintf_s(curShadowPath, bufferSizeInW - (curShadowPath - buffer), wcslen(redirectPrefix), L"%s", redirectPrefix);
			firstColon = TRUE;
		}
		origPtr++;
	}
	*curShadowPath = L'\0';

	return 0;
}


UNICODE_STRING MULTIP *ppi_convertWChart2UnicodeAllos(const wchar_t *path) {
	
	wchar_t *shadowPath = ppi_bm_path_malloc();
	if (shadowPath == NULL) {
		PPI_CRITICAL("Failed to allocated memory");
		return NULL;
	}

	UNICODE_STRING *shadowPathUnicode = (UNICODE_STRING *)ppi_bm_malloc(sizeof(UNICODE_STRING));
	if (shadowPathUnicode == NULL) {
		PPI_CRITICAL("Failed to allocated memory");
		ppi_bm_free(shadowPath);
		return NULL;
	}

	if (ppi_bm_copy(path, shadowPath) == -1) {
		wcscpy_s(shadowPath, PPI_PATH_WCHAR_SIZE, path);
	}
	shadowPathUnicode->Buffer = shadowPath;
	shadowPathUnicode->Length = shadowPathUnicode->MaximumLength = wcslen(shadowPath)*sizeof(wchar_t);

	return shadowPathUnicode;
}

UNICODE_STRING *PPI_replaceUserIDinRegAlloc(const wchar_t *originalReg, const wchar_t *toReplace) {
	
	if (originalReg == NULL)
		return nullptr;
	
	size_t extraPrefixBytes = 0;
	static int prefixPos = wcslen(L"\\REGISTRY\\USER\\" PPI_MULTIP_SYSTEM_SID_PREFIX);

	wchar_t *shadowPath = ppi_bm_path_malloc();
	if (shadowPath == NULL) {
		PPI_CRITICAL("Failed to allocated memory");
		return NULL;
	}

	UNICODE_STRING *shadowPathUnicode = (UNICODE_STRING *)ppi_bm_malloc(sizeof(UNICODE_STRING));
	if (shadowPathUnicode == NULL) {
		PPI_CRITICAL("Failed to allocated memory");
		ppi_bm_free(shadowPath);
		return NULL;
	}

	shadowPathUnicode->Buffer = shadowPath;
	const wchar_t *curOriginalPath = originalReg;
	wchar_t *curShadowPath = shadowPath;
	const wchar_t *curToReplace = toReplace;

	while (*curOriginalPath != NULL) {
		if ((curOriginalPath - originalReg < prefixPos) ||
			(curOriginalPath - originalReg >= (prefixPos + 4)))
			*curShadowPath = *curOriginalPath;
		else
			*curShadowPath = *curToReplace++;
		//PPI_CRITICAL("curShadowPath has %lc, %d, %d", *curShadowPath, curOriginalPath - originalReg, prefixPos);

		curShadowPath++;
		curOriginalPath++;
	}

	*curShadowPath = L'\0';

	shadowPathUnicode->Length = shadowPathUnicode->MaximumLength = wcslen(shadowPath)*sizeof(wchar_t);
	return shadowPathUnicode;
}


UNICODE_STRING MULTIP *ppi_convertUnicode2PrivateAlloc(const UNICODE_STRING *originalPath, const wchar_t *redirectPrefix) {

	PPI_BM_START_DETECT_BUFFER_CORRUPTION_P(9)
	wchar_t *fileName = originalPath->Buffer;
	size_t extraPrefixBytes = 0;
	/*
	if (wcsncmp(fileName, L"\\??\\", 4) == 0) {
		fileName += 4;
		extraPrefixBytes = 4;
	}
	*/
	wchar_t *shadowPath = ppi_bm_path_malloc();
	if (shadowPath == NULL) {
		PPI_CRITICAL("Failed to allocated memory");
		return NULL;
	}

	UNICODE_STRING *shadowPathUnicode = (UNICODE_STRING *)ppi_bm_malloc(sizeof(UNICODE_STRING));
	if (shadowPathUnicode == NULL) {
		PPI_CRITICAL("Failed to allocated memory");
		ppi_bm_free(shadowPath);
		return NULL;
	}

	PPI_BM_START_DETECT_BUFFER_CORRUPTION_P(8)
	shadowPathUnicode->Buffer = shadowPath;
	wchar_t *curShadowPath = shadowPath;

	BOOL firstColon = FALSE;
	unsigned int idx = 0;
	while (idx < originalPath->Length/sizeof(wchar_t)) {
		*curShadowPath = originalPath->Buffer[idx];
		curShadowPath++;
		if (!firstColon && originalPath->Buffer[idx] == L':') {
			curShadowPath += _snwprintf_s(curShadowPath, PPI_PATH_WCHAR_SIZE - (curShadowPath- shadowPath), wcslen(redirectPrefix), redirectPrefix);
			firstColon = TRUE;
		}
		idx++;
	}
	*curShadowPath = L'\0';

	PPI_BM_END_DETECT_BUFFER_CORRUPTION_P(8)
	shadowPathUnicode->Length = shadowPathUnicode->MaximumLength = wcslen(shadowPath)*sizeof(wchar_t);
	PPI_BM_END_DETECT_BUFFER_CORRUPTION_P(9)
	return shadowPathUnicode;
}


UNICODE_STRING MULTIP *ppi_convertUnicode2PrivateAlloc(const UNICODE_STRING *originalPath) {
	return ppi_convertUnicode2PrivateAlloc(originalPath, PPI_PRINCIPAL_MANAGER::getInstance().getCurrentPrincipal()->redirect_prefix);
}

void ppi_deallocUnicodeShadowPath(UNICODE_STRING *shadowPath) {
	
	if (shadowPath != NULL) {
		if (shadowPath->Buffer != NULL)
			ppi_bm_free(shadowPath->Buffer);
		ppi_bm_free(shadowPath);
	}
}


OBJECT_ATTRIBUTES *ppi_convertRegObjectAttributes2Benign(OBJECT_ATTRIBUTES *originalOA) {
	
	OBJECT_ATTRIBUTES *ShadowObjectAttributes = NULL;
	UNICODE_STRING *shadowPathUnicode = ppi_convertUnicode2BenignRegWAlloc(originalOA->ObjectName);
	if (!shadowPathUnicode) {
		//PPI_ERROR("Failed to convert path %ls to shadow path", originalOA->ObjectName->Buffer);
		return NULL;
	}
	ShadowObjectAttributes = (OBJECT_ATTRIBUTES *)ppi_bm_malloc(sizeof(OBJECT_ATTRIBUTES));
	if (ShadowObjectAttributes == NULL) {
		ppi_deallocUnicodeShadowPath(shadowPathUnicode);
		return NULL;
	}

	InitializeObjectAttributes(ShadowObjectAttributes, shadowPathUnicode, originalOA->Attributes, originalOA->RootDirectory, originalOA->SecurityDescriptor);
	ShadowObjectAttributes->SecurityQualityOfService = originalOA->SecurityQualityOfService;
	
	return ShadowObjectAttributes;
}


//MULTIP
OBJECT_ATTRIBUTES * MULTIP ppi_createObjectAttributesWithPrivatePath(const OBJECT_ATTRIBUTES *originalOA) {
	UNICODE_STRING *shadowPathUnicode = ppi_convertUnicode2PrivateAlloc(originalOA->ObjectName);
	if (!shadowPathUnicode) {
		PPI_ERROR("Failed to convert path %ls to shadow path", originalOA->ObjectName->Buffer);
		return NULL;
	}

	OBJECT_ATTRIBUTES *ShadowObjectAttributes = (OBJECT_ATTRIBUTES *)ppi_bm_malloc(sizeof(OBJECT_ATTRIBUTES));
	if (ShadowObjectAttributes == NULL) {
		PPI_CRITICAL("Failed to allocate memory for object attribute");
		ppi_deallocUnicodeShadowPath(shadowPathUnicode);
		return NULL;
	}

	InitializeObjectAttributes(ShadowObjectAttributes, shadowPathUnicode, originalOA->Attributes, originalOA->RootDirectory, originalOA->SecurityDescriptor);
	ShadowObjectAttributes->SecurityQualityOfService = originalOA->SecurityQualityOfService;

	return ShadowObjectAttributes;
}



void ppi_deallocObjectAttributes(OBJECT_ATTRIBUTES *oa) {
	if (oa != NULL) {
		if (oa->ObjectName != NULL)
			ppi_deallocUnicodeShadowPath(oa->ObjectName);
		ppi_bm_free(oa);
	}
}



OBJECT_ATTRIBUTES *PPI_OA::getPrivateOA() {

	if (_privateOA == NULL) {

		UNICODE_STRING *shadowPathUnicode = ppi_convertUnicode2PrivateAlloc(original_oa->ObjectName, _principal->redirect_prefix);
		if (!shadowPathUnicode) {
			PPI_ERROR("Failed to convert path %ls to shadow path", original_oa->ObjectName->Buffer);
			return NULL;
		}
		OBJECT_ATTRIBUTES *ShadowObjectAttributes = (OBJECT_ATTRIBUTES *)ppi_bm_malloc(sizeof(OBJECT_ATTRIBUTES));
		if (ShadowObjectAttributes == NULL) {
			PPI_CRITICAL("Failed to allocate memory for object attribute");
			ppi_deallocUnicodeShadowPath(shadowPathUnicode);
			return NULL;
		}
		InitializeObjectAttributes(ShadowObjectAttributes, shadowPathUnicode, original_oa->Attributes, original_oa->RootDirectory, original_oa->SecurityDescriptor);
		ShadowObjectAttributes->SecurityQualityOfService = original_oa->SecurityQualityOfService;

		ppi_recursivelyCreateDirectoriesIgnoreLastW(shadowPathUnicode->Buffer);

		_privateOA = ShadowObjectAttributes;
	}

	return this->_privateOA;
}

OBJECT_ATTRIBUTES *PPI_OA::getFirstReadableOA() {
	if (_firstReadableOA == NULL) {

		if (!ppi_isFullPathW(_printablePath))
			goto out;

		const list<const PPI_PRINCIPAL *> *listOfPrincipals = this->_principal->getTrustedPrincipals();

		if (listOfPrincipals->empty())
			goto out;

		UNICODE_STRING *shadowPathUnicode = NULL;
	
		BOOL fileExist;
		for (list<const PPI_PRINCIPAL *>::const_iterator it = listOfPrincipals->begin(); it != listOfPrincipals->end(); it++) {
			shadowPathUnicode = ppi_convertUnicode2PrivateAlloc(original_oa->ObjectName, (*it)->redirect_prefix);
			
			if ((ppi_utils_fileExist(shadowPathUnicode->Buffer, &fileExist) == 0) && fileExist) {
				
				if ((*it) == _principal) {
					if (_privateOA != NULL) {
						_firstReadableOA = _privateOA;
						ppi_deallocUnicodeShadowPath(shadowPathUnicode);
						goto out;
					}
				}
			
				OBJECT_ATTRIBUTES *ShadowObjectAttributes = (OBJECT_ATTRIBUTES *)ppi_bm_malloc(sizeof(OBJECT_ATTRIBUTES));
				if (ShadowObjectAttributes == NULL) {
					PPI_CRITICAL("Failed to allocate memory for object attribute");
					ppi_deallocUnicodeShadowPath(shadowPathUnicode);
					goto out;
				}

				InitializeObjectAttributes(ShadowObjectAttributes, shadowPathUnicode, original_oa->Attributes, original_oa->RootDirectory, original_oa->SecurityDescriptor);
				ShadowObjectAttributes->SecurityQualityOfService = original_oa->SecurityQualityOfService;

				_firstReadableOA = ShadowObjectAttributes;

				if ((*it) == _principal && _privateOA == NULL) {
					_privateOA = _firstReadableOA;
				}
				goto out;
			}
			ppi_deallocUnicodeShadowPath(shadowPathUnicode);
			shadowPathUnicode = NULL;
		}
	}
out:
	return _firstReadableOA;
}


bool PPI_OA::isExist() const {
	BOOL fileExist;
	if ((ppi_utils_fileExist(_printablePath, &fileExist) == 0) && fileExist)
		return true;
	return false;
}

void PPI_OA::_freeAllocatedOA(OBJECT_ATTRIBUTES *oa) {
	if (oa != NULL && oa != original_oa) {
		if (oa->ObjectName != NULL)
			ppi_deallocUnicodeShadowPath(oa->ObjectName);
		ppi_bm_free(oa);
	}
}


bool PPI_OA::copyFile2private() {
	if (getFirstReadableOA() == getPrivateOA())
		return true; //Files are already in the private location
	
	int rv;

	if (getPrivateOA()->ObjectName->Buffer == NULL) {
		PPI_ERROR("Destination path is NULL");
		return false;
	}
	
	if (!ppi_recursivelyCreateDirectoriesIgnoreLastW(getPrivateOA()->ObjectName->Buffer)) {
		PPI_ERROR("Failed to create directories for holding shadow file %ls", getPrivateOA()->ObjectName->Buffer);
		return false;
	}
		
	//Call with interception such that can ask helper on read if ACCESS_DENIED
	ppi_callWithInterception(CopyFileExW, rv, getFirstReadableOA()->ObjectName->Buffer, getPrivateOA()->ObjectName->Buffer, NULL, NULL, NULL, COPY_FILE_FAIL_IF_EXISTS);
	if (rv == 0) {
		PPI_ERROR("Failed to copy file from %ls to %ls due to %d", getFirstReadableOA()->ObjectName->Buffer, getPrivateOA()->ObjectName->Buffer, GetLastError());
		return false;
	}
	return true;
}

bool PPI_OA::isRedirectable() const {
	return ppi_OAIsRedirectable(original_oa);
}


PPI_DirectoryHandle::PPI_DirectoryHandle(HANDLE handle, const PPI_PRINCIPAL *principal): handle(handle), _principal(principal) {

	handlePath = ppi_bm_path_malloc();
	DWORD rv = GetFinalPathNameByHandle(handle, handlePath, PPI_PATH_WCHAR_SIZE, FILE_NAME_OPENED);
	if (rv == 0 || rv > PPI_PATH_WCHAR_SIZE) {
		PPI_NEWAPI_LOG("GetPath failed: %d", rv);
		ppi_bm_free(handlePath);
		handlePath = NULL;
	}
	
	const vector<wchar_t *> v = getAllDirectoryPath();
}

const wchar_t *PPI_DirectoryHandle::getHandlePath() const {
	return handlePath;
}

const vector<wchar_t *> &PPI_DirectoryHandle::getAllDirectoryPath() {
	
	if (allDirectoryPaths.empty()) {
		if (handlePath == NULL)
			return allDirectoryPaths;
		
		const list<const PPI_PRINCIPAL *> *listOfPrincipals = this->_principal->getTrustedPrincipals();

		wchar_t *undirectedHandlePath, *buffer2 = NULL;

		if (wcsstr(handlePath, this->_principal->redirect_prefix) == NULL)
			undirectedHandlePath = handlePath;
		else {
			buffer2 = ppi_bm_path_malloc();
			ppi_bm_copy(handlePath, buffer2);
			wchar_t *moveTo = wcsstr(buffer2, this->_principal->redirect_prefix);
			if (moveTo) {
				wchar_t *moveFrom = moveTo + wcslen(this->_principal->redirect_prefix);
				while (*moveFrom) {
					*moveTo = *moveFrom;
					moveTo++; moveFrom++;
				}
				*moveTo = L'\0';
			}
			undirectedHandlePath= buffer2;
		}
		BOOL fileExist = NULL;
		wchar_t *buffer3 = ppi_bm_path_malloc();
		for (list<const PPI_PRINCIPAL *>::const_iterator it = listOfPrincipals->begin(); it != listOfPrincipals->end(); it++) {
			if (ppi_insertRedirectPrefix(undirectedHandlePath, (*it)->redirect_prefix, buffer3, PPI_PATH_WCHAR_SIZE))
				continue;
			if ((ppi_utils_fileExist(buffer3, &fileExist) == 0) && fileExist) {
				allDirectoryPaths.push_back(buffer3);
				buffer3 = ppi_bm_path_malloc();
			}
		}
		ppi_bm_free(buffer3);
		if (buffer2)
			ppi_bm_free(buffer2);
	}
	return allDirectoryPaths;
}

PPI_DirectoryHandle::~PPI_DirectoryHandle() {
	if (handlePath != NULL) {
		ppi_bm_free(handlePath);
		handlePath = NULL;
	}

	vector<wchar_t *>::iterator it;
	for (it = allDirectoryPaths.begin(); it != allDirectoryPaths.end(); it++) {
		ppi_bm_free(*it);
	}
}

