//
// Portable Integrity Protection (PIP) System -
// Copyright (C) 2015 Secure Systems Laboratory, Stony Brook University
//
// This file is part of Portable Integrity Protection (PIP) System.
//
// Portable Integrity Protection (PIP) System is free software: you can
// redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Portable Integrity Protection (PIP) System is distributed in the hope that it
// will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Portable Integrity Protection (PIP) System.  If not, see
// <http://www.gnu.org/licenses/>.
//

#include "ppi_debug.h"
#include "ppi_utils.h"
#include "ppi_config.h"
#include <Shlwapi.h>
#include <stdio.h>
#include <Sddl.h>
#include <Aclapi.h>

#include "ppi_version.h"
#include "ppi_bufferManager.h"

#pragma comment (lib, "Advapi32.lib")

#define d (256*256)
#define q 32767


wchar_t *untrustedSID = PPI_UNTRUSTED_SIDSTRING;
size_t untrustedSIDSize = wcslen(untrustedSID);
int untrustedSIDHash;
BOOL untrustedSIDHashCalculated = FALSE;
int untrustedSID_h;

const wchar_t *ppi_firstUntrustedSID(const wchar_t *string, size_t n) {
	
	size_t count, temp;;
	int stringHash = 0;

	if (!untrustedSIDHashCalculated) {
		int h = 1;
		untrustedSIDHash = 0;
				
		for (count = 0; count < untrustedSIDSize-1; count++)
			h = (h*d)%q;
	
		for (count = 0; count < untrustedSIDSize; count++) {
			untrustedSIDHash = (d*untrustedSIDHash + untrustedSID[count]) % q;
		}
		untrustedSIDHashCalculated = TRUE;
		untrustedSID_h = h;
	}

	if (n < untrustedSIDSize)
		return NULL;
	
	for (count = 0; count < untrustedSIDSize; count++) {
		stringHash = (d*stringHash + string[count]) % q;
	}

	for (count=0; count<=n-untrustedSIDSize; count++) {
		if (untrustedSIDHash == stringHash) {
			for (temp=0; temp<untrustedSIDSize; temp++) {
				if (string[count+temp] != untrustedSID[temp])
					break;
			}
			if (temp == untrustedSIDSize)
				return string + count;
		}
		if (count < n-untrustedSIDSize) {
			stringHash = (d*(stringHash - string[count]*untrustedSID_h) + string[count+untrustedSIDSize])%q;
			if (stringHash < 0)
				stringHash += q;
		}
	}
	return NULL;
}



wchar_t *ppi_utils_programPath = NULL;

const wchar_t *ppi_utils_getProgramPath() {
	if (ppi_utils_programPath == NULL) {
		/* Do Not rely on bufferManager Here!!!: Debug_logging will not work! */
		wchar_t *tempPath = (wchar_t*)malloc(PPI_PATH_WCHAR_SIZE*sizeof(wchar_t));
		if (GetModuleFileName(NULL, tempPath, PPI_PATH_WCHAR_SIZE) == 0 || ppi_utils_programPath != NULL) {
			free(tempPath);
		} else {
			ppi_utils_programPath = tempPath;
		}
	}
	return ppi_utils_programPath;
}



BOOL ppi_utils_intInArray(int toFind, const int intArray[]) {
	int count = 0;
	while (intArray[count] != -1) {
		if (toFind == intArray[count])
			return TRUE;
		count++;
	}
	return FALSE;
}

BOOL ppi_utils_wstringInArray(const wchar_t *toFind, const wchar_t *stringArray[]) {
	int count=0;
	while (stringArray[count] != NULL) {
		if (_wcsicmp(toFind, stringArray[count]) == 0)
			return TRUE;
		count++;
	}
	return FALSE;
}

//Case insensitive
BOOL ppi_utils_wstringStartWithArray(const wchar_t *toFind, const wchar_t *stringArray[]) {
	int count=0;
	while (stringArray[count] != NULL) {
		if (_wcsnicmp(toFind, stringArray[count], wcslen(stringArray[count])) == 0)
			return true;
		count++;
	}
	return false;
}


int ppi_utils_fileExist(LPCWSTR fileName, BOOL *fileExists) {
	int rv = -1;
	
	if (wcsncmp(fileName, L"\\??\\", 4) == 0)
		fileName += 4;
	
	if (INVALID_FILE_ATTRIBUTES != GetFileAttributesW(fileName)) {
		*fileExists = true;
		rv = 0;
		goto out;
	} else {
		DWORD error = GetLastError();
		if ((error == ERROR_FILE_NOT_FOUND) || (error == ERROR_PATH_NOT_FOUND) || (error == ERROR_INVALID_NAME)) {
			*fileExists = false;
			rv = 0;
			goto out;
		} else {
			PPI_ERROR("Cannot check if file %ls exists: error: %d", fileName, GetLastError());
			goto out;
		}
	}
out:
	return rv;
}

//TODO:
int ppi_utils_regExist(const wchar_t *regPath, BOOL *regExists) {
	int rv = 0;




	return rv;
}


int ppi_utils_convertChar2Wchar(const char *input, wchar_t *buffer, size_t buflen) {
	
	if (buffer != NULL) {
		if (input == NULL && buflen > 1) {
			buffer[0] = L'\0';
			return 0;
		}
	
		if (0 == MultiByteToWideChar(CP_ACP, 0, input, -1, buffer, buflen)) {
			PPI_ERROR("Failed to convert string from A to W on input %s", input);
			return -1;
		} else {
			return 0;
		}
	}
	return -1;
}

void ppi_utils_grantUntrustedReadAccess(LPWSTR filename) {
	
	//TODO: Add ace to untrusted files created
	bool result = false;
	PACL pACL = NULL, pNewAcl = NULL;
	PSECURITY_DESCRIPTOR pSD = NULL, psdNew = NULL;
	ACL_SIZE_INFORMATION aclSizeInfo;
	PEXPLICIT_ACCESS pListOfExplicitEntries = NULL;
	PVOID pTempAce;

	DWORD dwNewAclSize;
	PSID psid = NULL;

	unsigned int i;
	

	DWORD dwRes = GetNamedSecurityInfo(filename, 
		SE_FILE_OBJECT, DACL_SECURITY_INFORMATION|OWNER_SECURITY_INFORMATION|GROUP_SECURITY_INFORMATION, NULL, NULL, &pACL, NULL, &pSD);
		
	ULONG countOfExplicitEntries = 0;

	if (ERROR_SUCCESS != dwRes) {
		PPI_ERROR("Failed at GetNamedSecurityInfo, errno %u %u\n", dwRes, GetLastError());
	} else {
		if (pACL == NULL)
			goto cleanup;

		if (!GetAclInformation(pACL, (LPVOID)&aclSizeInfo, sizeof(ACL_SIZE_INFORMATION), AclSizeInformation))
			goto cleanup;

		if (!ConvertStringSidToSid(PPI_UNTRUSTED_SIDSTRING, &psid)) {
			PPI_ERROR("Failed to convert string to sid\n");
			goto cleanup;
		}
		
		dwNewAclSize = aclSizeInfo.AclBytesInUse + sizeof(ACCESS_ALLOWED_ACE) + GetLengthSid(psid) - sizeof(DWORD);

		pNewAcl = (PACL)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, dwNewAclSize);
		if (pNewAcl == NULL)
			goto cleanup;

		if (!InitializeAcl(pNewAcl, dwNewAclSize, ACL_REVISION))
			goto cleanup;

		if (aclSizeInfo.AceCount) {
			for (i=0; i< aclSizeInfo.AceCount; i++) {
				if (!GetAce(pACL, i, &pTempAce))
					goto cleanup;
				if (!AddAce(pNewAcl, ACL_REVISION, MAXDWORD, pTempAce, ((PACE_HEADER)pTempAce)->AceSize))
					goto cleanup;
			}
		}
		if (!AddAccessAllowedAce(pNewAcl, ACL_REVISION, FILE_GENERIC_READ, psid)) {
			PPI_ERROR("Failed at AddAccessAllowedAce Read\n");
			goto cleanup;
		}

		if (ERROR_SUCCESS != SetNamedSecurityInfo(filename, SE_FILE_OBJECT, DACL_SECURITY_INFORMATION, NULL, NULL, pNewAcl, NULL)) {
			PPI_ERROR("Failed to set new acl\n");
			goto cleanup;
		}

		result = true;
	}
cleanup:
	if (pSD != NULL)
		LocalFree((HLOCAL)pSD);
	if (psdNew != NULL)
		LocalFree((HLOCAL)psdNew);
	if (psid != NULL)
		LocalFree((HLOCAL)psid);
	if (pNewAcl != NULL)
		HeapFree(GetProcessHeap(), 0, pNewAcl);

}


