/****************************** Module Header ******************************\
* Module Name:  SampleService.cpp
* Project:      CppWindowsService
* Copyright (c) Microsoft Corporation.
* 
* Provides a sample service class that derives from the service base class - 
* CServiceBase. The sample service logs the service start and stop 
* information to the Application event log, and shows how to run the main 
* function of the service in a thread pool worker thread.
* 
* This source is subject to the Microsoft Public License.
* See http://www.microsoft.com/en-us/openness/resources/licenses.aspx#MPL.
* All other rights reserved.
* 
* THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, 
* EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED 
* WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.
\***************************************************************************/


#include "ppi_packet.h"
#include <winsock2.h>
#include <ws2tcpip.h>


#include "ppi_config.h"
#include "ppi_helper.h"
#include "ppi_utils.h"
#include "ppi_debug.h"

#include "ppi_level.h"

#include "ppi_version.h"

#pragma region Includes
#include "SampleService.h"
#include "ThreadPool.h"
#pragma endregion


#pragma comment (lib, "Ws2_32.lib")
#pragma comment (lib, "Mswsock.lib")
#pragma comment (lib, "AdvApi32.lib")



#define DEFAULT_BUFLEN 512


CSampleService::CSampleService(PWSTR pszServiceName, 
                               BOOL fCanStop, 
                               BOOL fCanShutdown, 
                               BOOL fCanPauseContinue)
: CServiceBase(pszServiceName, fCanStop, fCanShutdown, fCanPauseContinue)
{
    m_fStopping = FALSE;

    // Create a manual-reset event that is not signaled at first to indicate 
    // the stopped signal of the service.
    m_hStoppedEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
    if (m_hStoppedEvent == NULL)
    {
        throw GetLastError();
    }
}


CSampleService::~CSampleService(void)
{
    if (m_hStoppedEvent)
    {
        CloseHandle(m_hStoppedEvent);
        m_hStoppedEvent = NULL;
    }
}


//
//   FUNCTION: CSampleService::OnStart(DWORD, LPWSTR *)
//
//   PURPOSE: The function is executed when a Start command is sent to the 
//   service by the SCM or when the operating system starts (for a service 
//   that starts automatically). It specifies actions to take when the 
//   service starts. In this code sample, OnStart logs a service-start 
//   message to the Application log, and queues the main service function for 
//   execution in a thread pool worker thread.
//
//   PARAMETERS:
//   * dwArgc   - number of command line arguments
//   * lpszArgv - array of command line arguments
//
//   NOTE: A service application is designed to be long running. Therefore, 
//   it usually polls or monitors something in the system. The monitoring is 
//   set up in the OnStart method. However, OnStart does not actually do the 
//   monitoring. The OnStart method must return to the operating system after 
//   the service's operation has begun. It must not loop forever or block. To 
//   set up a simple monitoring mechanism, one general solution is to create 
//   a timer in OnStart. The timer would then raise events in your code 
//   periodically, at which time your service could do its monitoring. The 
//   other solution is to spawn a new thread to perform the main service 
//   functions, which is demonstrated in this code sample.
//
void CSampleService::OnStart(DWORD dwArgc, LPWSTR *lpszArgv)
{
    // Log a service start message to the Application log.
    WriteEventLogEntry(L"CppWindowsService in OnStart", 
        EVENTLOG_INFORMATION_TYPE);
	
	PPI_INFO("Service started");

    // Queue the main service function for execution in a worker thread.
    CThreadPool::QueueUserWorkItem(&CSampleService::ServiceWorkerThread, this);
}

void HandleRequest2(void *socket);



//
//   FUNCTION: CSampleService::ServiceWorkerThread(void)
//
//   PURPOSE: The method performs the main function of the service. It runs 
//   on a thread pool worker thread.
//
void CSampleService::ServiceWorkerThread(void)
{


    WSADATA wsaData;
    int iResult;

    SOCKET ListenSocket = INVALID_SOCKET;
    SOCKET ClientSocket = INVALID_SOCKET;

    struct addrinfo *result = NULL;
    struct addrinfo hints;
	ULONG NonBlock;

    int iSendResult;
    char recvbuf[DEFAULT_BUFLEN];
    int recvbuflen = DEFAULT_BUFLEN;
    
    // Initialize Winsock
    iResult = WSAStartup(MAKEWORD(2,2), &wsaData);
    if (iResult != 0) {
        printf("WSAStartup failed with error: %d\n", iResult);
		throw GetLastError();
    }

    ZeroMemory(&hints, sizeof(hints));
    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;
    hints.ai_flags = AI_PASSIVE;

    // Resolve the server address and port
    iResult = getaddrinfo(NULL, DEFAULT_PORT, &hints, &result);
    if ( iResult != 0 ) {
        printf("getaddrinfo failed with error: %d\n", iResult);
        WSACleanup();
		throw GetLastError();
    }

    // Create a SOCKET for connecting to server
    ListenSocket = socket(result->ai_family, result->ai_socktype, result->ai_protocol);
    if (ListenSocket == INVALID_SOCKET) {
        printf("socket failed with error: %ld\n", WSAGetLastError());
        freeaddrinfo(result);
        WSACleanup();
		throw GetLastError();
    }

    // Setup the TCP listening socket
    iResult = bind( ListenSocket, result->ai_addr, (int)result->ai_addrlen);
    if (iResult == SOCKET_ERROR) {
        printf("bind failed with error: %d\n", WSAGetLastError());
        freeaddrinfo(result);
        closesocket(ListenSocket);
        WSACleanup();
		throw GetLastError();
    }

    freeaddrinfo(result);

    iResult = listen(ListenSocket, SOMAXCONN);
    if (iResult == SOCKET_ERROR) {
        printf("listen failed with error: %d\n", WSAGetLastError());
        closesocket(ListenSocket);
        WSACleanup();
		throw GetLastError();
    }

	NonBlock = 1;
    if (ioctlsocket(ListenSocket, FIONBIO, &NonBlock) == SOCKET_ERROR) {
		PPI_ERROR("Cannot set listen socket to nonblock");
	}

	
	DWORD currentPid = GetCurrentProcessId();
	PPI_INFO("Current pid is %u\n", currentPid);

	this->ListenSocket = ListenSocket;
	
    // Periodically check if the service is stopping.
    while (!m_fStopping)
    {

		FD_SET Reader;

		struct timeval tv;
		tv.tv_sec = 10; 
		tv.tv_usec = 0;
		int Total;
		FD_ZERO(&Reader);
        // Check for connection attempts.
        FD_SET(ListenSocket, &Reader);

		if ((Total = select(0, &Reader, NULL, NULL, &tv)) == SOCKET_ERROR)
        {
             PPI_ERROR("Winsock tutorial error: select function returned with error %d\n", WSAGetLastError());
		}

		//PPI_CRITICAL("Time expired? TOTAL: %d", Total);
		if (Total == 0)
			continue;
		

		if (FD_ISSET(ListenSocket, &Reader)) {
			PPI_INFO("Accept new client!");
			HandleClient *hc = new HandleClient(accept(this->ListenSocket, NULL, NULL));
			CThreadPool::QueueUserWorkItem(&HandleClient::AcceptNewClient, hc);
			
		} else {
			PPI_CRITICAL("Reader is not set");
		}
		
    }
	
    // No longer need server socket
    closesocket(ListenSocket);

	WSACleanup();
    // Signal the stopped event.
    SetEvent(m_hStoppedEvent);
}



HandleClient::HandleClient(SOCKET socket) {
	this->clientSocket = socket;
}

HandleClient::~HandleClient(void)
{
    //CloseHandle(this->clientSocket);
}

void HandleClient::AcceptNewClient(void) {

	SOCKET ClientSocket = this->clientSocket;

	int iResult;

	char message[4096*10];

	int pkt_size;
	int cmd_no;
	DWORD pid;
	HANDLE peerProcess = INVALID_HANDLE_VALUE;

	if (ClientSocket == INVALID_SOCKET) {
		//ppi_log("Client socket is invalid");
		return;
	}

	u_long NonBlock = 0;
    if (ioctlsocket(ClientSocket, FIONBIO, &NonBlock) == SOCKET_ERROR) {
		PPI_ERROR("Cannot set socket to blocking mode");
	}

	iResult = recv(ClientSocket, message, sizeof(int)*2 + sizeof(DWORD), MSG_PEEK);
	if (iResult != sizeof(int)*2 + sizeof(DWORD))
		goto out;

	pid = *((DWORD *)(message + sizeof(int)*2));
	peerProcess = OpenProcess(PROCESS_DUP_HANDLE, false, pid);
	PPI_INFO("Received peer: pid: %d", pid);

	while (1) {
		
		iResult = recv(ClientSocket, message, sizeof(int)*2 + sizeof(DWORD), MSG_PEEK);
		//ppi_log("Message received");

		if (iResult != sizeof(int)*2 + sizeof(DWORD)) {
			PPI_ERROR("recv is not expected: %d", iResult);
			break;
		}
		pkt_size = ((int *)message)[0];
		cmd_no = ((int *)message)[1];
		
		//PPI_CRITICAL("Request for cmd_no: %d", cmd_no);	

		if (!ppi_pkt_recvCorrectly(ClientSocket, message, pkt_size)) {
			PPI_CRITICAL("Failed to receive packet correctly from client");
			break;
		}

#define ppi_helper_call(call) process_ ##call((struct ppi_pkt_ ##call *) &message, &response, pid, peerProcess)

#define ppi_helper_case(_call) \
	case SYS_ ##_call: { \
		struct ppi_pkt_ ##_call ##_response response = ppi_getPktResponse(_call); \
		ppi_helper_call(_call); \
		if (cmd_no != response.l_cmdno) \
			PPI_ERROR("Packet cmd_no and response not match!!!: received: %d, generated: %d", cmd_no, response.l_cmdno);\
		if (!SEND_PKT_CORRECT(ClientSocket, response)) { \
			PPI_ERROR("Failed to send response"); \
			goto out; \
		} \
		break; \
	}

		switch (cmd_no) {

			ppi_helper_case(NtCreateFile);
			ppi_helper_case(NtOpenDirectoryObject);
			ppi_helper_case(NtSetInformationFile);
			ppi_helper_case(NtQueryAttributesFile);

#ifdef WIN81
			ppi_helper_case(NtQueryFullAttributesFile);
#endif
			ppi_helper_case(NtOpenKeyEx);
			ppi_helper_case(NtClose);
			ppi_helper_case(DupToken);

		default:
			PPI_ERROR("Unexpected call: %d", cmd_no);
			goto out;
		}
	}

out:
	closesocket(ClientSocket);

	if (peerProcess != INVALID_HANDLE_VALUE)
		CloseHandle(peerProcess);

	delete this;

	return;

}


//
//   FUNCTION: CSampleService::OnStop(void)
//
//   PURPOSE: The function is executed when a Stop command is sent to the 
//   service by SCM. It specifies actions to take when a service stops 
//   running. In this code sample, OnStop logs a service-stop message to the 
//   Application log, and waits for the finish of the main service function.
//
//   COMMENTS:
//   Be sure to periodically call ReportServiceStatus() with 
//   SERVICE_STOP_PENDING if the procedure is going to take long time. 
//
void CSampleService::OnStop()
{
    // Log a service stop message to the Application log.
    WriteEventLogEntry(L"CppWindowsService in OnStop", 
        EVENTLOG_INFORMATION_TYPE);

    // Indicate that the service is stopping and wait for the finish of the 
    // main service function (ServiceWorkerThread).
    m_fStopping = TRUE;
    if (WaitForSingleObject(m_hStoppedEvent, INFINITE) != WAIT_OBJECT_0)
    {
        throw GetLastError();
    }
}