//
// Portable Integrity Protection (PIP) System -
// Copyright (C) 2015 Secure Systems Laboratory, Stony Brook University
//
// This file is part of Portable Integrity Protection (PIP) System.
//
// Portable Integrity Protection (PIP) System is free software: you can
// redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Portable Integrity Protection (PIP) System is distributed in the hope that it
// will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Portable Integrity Protection (PIP) System.  If not, see
// <http://www.gnu.org/licenses/>.
//

#include "ppi_calls.h"

#include <windows.h>
#include <stdio.h>
#include "detours.h"

#include "ppi_debug.h"
#include "ppi_version.h"

#include "ppi_windowsIntegrity.h"
#include "ppi_principal.h"

#include "ppi_level.h"

//////////////////////////////////////////////////////////////////////////////
#pragma warning(disable:4127)   // Many of our asserts are constants.

#define ASSERT_ALWAYS(x)   \
    do {                                                        \
    if (!(x)) {                                                 \
            AssertMessage(#x, __FILE__, __LINE__);              \
            DebugBreak();                                       \
    }                                                           \
    } while (0)

#ifndef NDEBUG
#define ASSERT(x)           ASSERT_ALWAYS(x)
#else
#define ASSERT(x)
#endif

#define UNUSED(c)       (c) = (c)

//#define DO_NOT_INTERCEPT

static HMODULE s_hInst = NULL;
static CHAR s_szDllPath[MAX_PATH];


HMODULE __declspec( dllexport ) WINAPI aaa() {
	return s_hInst;
}



const wchar_t *ppi_ignoreProcesses[] = {
	L"C:\\WINDOWS\\system32\\VBoxTray.exe",
	L"c:\\Program Files\\Microsoft Visual Studio 10.0\\VC\\bin\\CL.exe",
	L"c:\\Program Files\\Microsoft Visual Studio 10.0\\VC\\vcpackages\\VCPkgSrv.exe",
	L"C:\\pip-windows\\pip_daemon\\Debug\\pip_daemon.exe",
	L"C:\\Windows\\system32\\SearchProtocolHost.exe",
	L"C:\\Windows\\system32\\wininit.exe", // AccessCheck hang???
	L"C:\\Windows\\system32\\msiexec.exe", //AccessCheck returns wrong information
	NULL};

const wchar_t *ppi_customTestProcesses[] = {
	L"C:\\pip-windows\\ppi_detour2\\Debug\\tester.exe",
	NULL};

#define ppi_assign_originalFunc(funcName, library) \
	P_ ##funcName Original ##funcName = (P_ ##funcName)::GetProcAddress(::GetModuleHandle(library), #funcName)

#define ppi_assign_originalFuncNew(funcName, library) \
	P_ ##funcName Original ##funcName ##New = (P_ ##funcName)::GetProcAddress(::GetModuleHandle(library), #funcName)


ppi_assign_originalFuncNew(NtCreateFile, L"ntdll");
ppi_assign_originalFuncNew(NtOpenFile, L"ntdll");
ppi_assign_originalFuncNew(NtQueryDirectoryFile, L"ntdll");
ppi_assign_originalFuncNew(NtClose, L"ntdll");

ppi_assign_originalFuncNew(CreateProcessW, L"kernel32");
ppi_assign_originalFuncNew(CreateProcessA, L"kernel32");
ppi_assign_originalFuncNew(NtOpenKey, L"ntdll");


ppi_assign_originalFunc(CreateFileW, L"kernel32");
ppi_assign_originalFunc(NtCreateFile, L"ntdll");
ppi_assign_originalFunc(NtOpenFile, L"ntdll");
ppi_assign_originalFunc(NtClose, L"ntdll");
ppi_assign_originalFunc(NtOpenDirectoryObject, L"ntdll");
ppi_assign_originalFunc(NtQueryDirectoryFile, L"ntdll");
ppi_assign_originalFunc(NtSetInformationFile, L"ntdll");
ppi_assign_originalFunc(NtQueryAttributesFile, L"ntdll");
ppi_assign_originalFunc(NtCreateKey, L"ntdll");
ppi_assign_originalFunc(NtOpenKey, L"ntdll");
ppi_assign_originalFunc(NtOpenKeyEx, L"ntdll");
ppi_assign_originalFunc(NtOpenKeyTransacted, L"ntdll");
ppi_assign_originalFunc(NtOpenKeyTransactedEx, L"ntdll");
ppi_assign_originalFunc(NtSetValueKey, L"ntdll");

ppi_assign_originalFunc(MoveFileExW, L"kernel32");
ppi_assign_originalFunc(CreateProcessW, L"kernel32");
ppi_assign_originalFunc(CreateProcessA, L"kernel32");

#ifdef WIN81
ppi_assign_originalFunc(NtQueryFullAttributesFile, L"ntdll");
#endif


DWORD ppi_dwTlsIndex;


struct ppi_tls *ppi_getTls() {
	struct ppi_tls *tls = (struct ppi_tls *)TlsGetValue(ppi_dwTlsIndex);
	if (tls == NULL)
		tls = ppi_allocateTls();
	return tls;
}

void ppi_destoryTls() {
	struct ppi_tls *lpvData = (struct ppi_tls *)TlsGetValue(ppi_dwTlsIndex);
	if (lpvData != NULL)
		LocalFree((HLOCAL) lpvData);
}

struct ppi_tls *ppi_allocateTls() {
	struct ppi_tls *lpvData = (struct ppi_tls *) LocalAlloc(LPTR, sizeof(struct ppi_tls));
	if (lpvData != NULL) {
		TlsSetValue(ppi_dwTlsIndex, lpvData);
		lpvData->ppi_isIntercepted = false;
	}
	return lpvData;
}


PCHAR DetRealName(PCHAR psz)
{
    PCHAR pszBeg = psz;
    // Move to end of name.
    while (*psz) {
        psz++;
    }
    // Move back through A-Za-z0-9 names.
    while (psz > pszBeg &&
           ((psz[-1] >= 'A' && psz[-1] <= 'Z') ||
            (psz[-1] >= 'a' && psz[-1] <= 'z') ||
            (psz[-1] >= '0' && psz[-1] <= '9'))) {
        psz--;
    }
    return psz;
}

VOID DetAttach(PVOID *ppbReal, PVOID pbMine, PCHAR psz)
{
    LONG l = DetourAttach(ppbReal, pbMine);
    if (l != 0) {
        PPI_ERROR("Attach failed: `%s': error %d\n", psz, l);
    }
}

VOID DetDetach(PVOID *ppbReal, PVOID pbMine, PCHAR psz)
{
    LONG l = DetourDetach(ppbReal, pbMine);
    if (l != 0) {
        PPI_ERROR("Detach failed: `%s': error %d\n", psz, l);
    }
}

#define ATTACH(x)       DetAttach(&(PVOID&)Original##x, Hooked##x, #x)
#define DETACH(x)       DetDetach(&(PVOID&)Original##x, Hooked##x, #x)

LONG AttachDetours(VOID)
{

#ifdef DO_NOT_INTERCEPT
	return 0;
#endif

    DetourTransactionBegin();
    DetourUpdateThread(GetCurrentThread());
	
	//ATTACH(NtOpenKeyNew);
	
	ATTACH(NtCreateFileNew);
	ATTACH(NtOpenFileNew);
	ATTACH(NtQueryDirectoryFileNew);
	
	ATTACH(NtSetInformationFile);
	ATTACH(CreateProcessANew);
	ATTACH(CreateProcessWNew);
	
	ATTACH(NtQueryAttributesFile);
	//ATTACH(NtCloseNew);

#ifdef WIN81
	ATTACH(NtQueryFullAttributesFile);
#endif
	
	
	LONG rv = DetourTransactionCommit();
    return rv;
}

LONG DetachDetours(VOID)
{
#ifdef DO_NOT_INTERCEPT
	return 0;
#endif

    DetourTransactionBegin();
    DetourUpdateThread(GetCurrentThread());
	
	//DETACH(NtOpenKeyNew);

	DETACH(NtCreateFileNew);
	DETACH(NtOpenFileNew);
	DETACH(NtQueryDirectoryFileNew);
	
	
	DETACH(NtSetInformationFile);
	DETACH(CreateProcessWNew);
	DETACH(CreateProcessANew);
	
	//DETACH(NtCloseNew);

	DETACH(NtQueryAttributesFile);
#ifdef WIN81
	DETACH(NtQueryFullAttributesFile);
#endif
	

    return DetourTransactionCommit();
}
//
//////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////// Logging System.
//
static BOOL s_bLog = 1;
static LONG s_nTlsIndent = -1;
static LONG s_nTlsThread = -1;
static LONG s_nThreadCnt = 0;

//////////////////////////////////////////////////////////////////////////////
//
// DLL module information
//
BOOL ThreadAttach(HMODULE hDll)
{
    (void)hDll;
	
	ppi_allocateTls();

    return TRUE;
}

BOOL ThreadDetach(HMODULE hDll)
{
    (void)hDll;
	
	ppi_destoryTls();

    return TRUE;
}

BOOL ProcessAttach(HMODULE hDll)
{
	struct ppi_tls *tls = NULL;
    s_bLog = FALSE;
    s_nTlsIndent = TlsAlloc();
    s_nTlsThread = TlsAlloc();
	
	if ((ppi_dwTlsIndex = TlsAlloc()) == TLS_OUT_OF_INDEXES) {
		PPI_ERROR("Failed to allocate TLS for main process");
	} else {
		tls = ppi_allocateTls();
	}
	
	if (tls != NULL)
		tls->ppi_isIntercepted = TRUE;

    s_hInst = hDll;
    GetModuleFileNameA(s_hInst, s_szDllPath, ARRAYSIZE(s_szDllPath));
	
    ThreadAttach(hDll);
    s_bLog = TRUE;

    LONG error = AttachDetours();
    if (error != NO_ERROR) {
        PPI_ERROR("### Error attaching detours: %d", error);
    }
	

	if (tls != NULL)
		tls->ppi_isIntercepted = FALSE;

    return TRUE;
}

BOOL ProcessDetach(HMODULE hDll)
{
    ThreadDetach(hDll);
    s_bLog = FALSE;

    LONG error = DetachDetours();
    if (error != NO_ERROR) {
		PPI_ERROR("### Error detaching detours: %d", error);
    }


	//ppi_destoryTls();
	TlsFree(ppi_dwTlsIndex);

    return TRUE;
}



void ppi_customTest() {
}

BOOL ppi_isIgnoredProcess() {
	
	if (ppi_utils_wstringInArray(ppi_utils_getProgramPath(), ppi_ignoreProcesses)) {
		return TRUE;
	}
	
	if (ppi_windowsIntegrity_getProcessIntegrityLevel() < SECURITY_MANDATORY_MEDIUM_RID)
		return TRUE;
	//PPI_CRITICAL("Intercepting calls from %ls", ppi_utils_getProgramPath());
	
	if (ppi_utils_wstringInArray(ppi_utils_getProgramPath(), ppi_customTestProcesses)) {
		ppi_customTest();
		return TRUE;
	}
	return FALSE;
}


BOOL APIENTRY DllMain(HINSTANCE hModule, DWORD dwReason, PVOID lpReserved)
{
    (void)hModule;
    (void)lpReserved;

	if (DetourIsHelperProcess()) {
        return TRUE;
    }	
    if (ppi_isIgnoredProcess()) {
		return TRUE;
	}

    switch (dwReason) {
      case DLL_PROCESS_ATTACH:
        DetourRestoreAfterWith();
        return ProcessAttach(hModule);
      case DLL_PROCESS_DETACH:
        return ProcessDetach(hModule);
      case DLL_THREAD_ATTACH:
        return ThreadAttach(hModule);
      case DLL_THREAD_DETACH:
        return ThreadDetach(hModule);
    }
    return TRUE;
}
//
///////////////////////////////////////////////////////////////// End of File.
