//
// Portable Integrity Protection (PIP) System -
// Copyright (C) 2015 Secure Systems Laboratory, Stony Brook University
//
// This file is part of Portable Integrity Protection (PIP) System.
//
// Portable Integrity Protection (PIP) System is free software: you can
// redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Portable Integrity Protection (PIP) System is distributed in the hope that it
// will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Portable Integrity Protection (PIP) System.  If not, see
// <http://www.gnu.org/licenses/>.
//

#include "ppi_calls.h"
#include "ppi_debug.h"
#include "ppi_main.h"
#include "ppi_level.h"
#include "ppi_utils.h"
#include <list>
#include <algorithm>


class PPI_CommandLine {
public:
	PPI_CommandLine(const wchar_t *applicationName, const wchar_t *commandLine): applicationName_(applicationName), commandLine_(commandLine) {
		applicationBuf_ = NULL;
		modifiedCommandLine_ = NULL;
	};

	const wchar_t *getApplicationPath() {
		if (applicationName_ != NULL)
			return applicationName_;

		if (applicationBuf_ == NULL)
			parseCommand();
		return applicationBuf_;
	};

	const list<wchar_t *> &getArguments() {
		if (arguments_.empty())
			parseCommand();
		return arguments_;
	}

	const list<wchar_t *> &getFileArguments() {
		if (fileLikeArguments_.empty()) {
			getArguments();
			for (list<wchar_t *>::iterator it=arguments_.begin(); it !=arguments_.end(); it++) {
				BOOL fileExists;
				if (ppi_utils_fileExist(*it, &fileExists) == 0 && fileExists)
					fileLikeArguments_.push_back(*it);
			}
		}
		return fileLikeArguments_;
	}

	~PPI_CommandLine() {
		for (list<wchar_t *>::iterator it=arguments_.begin(); it !=arguments_.end(); it++) {
			ppi_bm_free(*it);
		}

	};

	wchar_t *getRunAsCommand(const PPI_PRINCIPAL *p) {
		if (modifiedCommandLine_ == NULL)
			modifiedCommandLine_ = ppi_bm_path_malloc();
		
		size_t printedCount = 0;
		printedCount += _snwprintf_s(modifiedCommandLine_, PPI_PATH_WCHAR_SIZE, PPI_PATH_WCHAR_SIZE, L"runas /savecred /env /user:%ls \"", p->username);

		wchar_t *cur = modifiedCommandLine_ + printedCount;
		for (const wchar_t *p = commandLine_; *p != NULL; p++) {
			if (*p == L'"') {
				*cur = L'\\';
				cur++;
			}
			*cur = *p;
			cur++;
		}
		*cur = L'"';
		cur++;
		*cur = L'\0';

		return modifiedCommandLine_;
	}

private:
	const wchar_t *applicationName_;
	const wchar_t *commandLine_;
	wchar_t *applicationBuf_;
	list<wchar_t *> arguments_;
	list<wchar_t *> fileLikeArguments_;
	wchar_t *modifiedCommandLine_;

	void parseCommand() {
		
		if (commandLine_ == NULL)
			return;

		enum STATE {INQUOTE, NONQUOTE};
		STATE s = NONQUOTE;

#define END_CURRENT_AND_CREATE_NEW \
		do {\
			if (curFillingIdx == 0) \
				break; \
			curArg[curFillingIdx++] = L'\0'; \
			curArg = ppi_bm_path_malloc(); \
			arguments_.push_back(curArg); \
			curFillingIdx = 0; \
			curArg[curFillingIdx] = L'\0'; \
		} while (0)

		int idx = 0;
		int curFillingIdx = 0;
		wchar_t *curArg = ppi_bm_path_malloc();
		arguments_.push_back(curArg);
		int quoteCount = 0;
		while (commandLine_[idx] != L'\0') {
			switch(commandLine_[idx]) {
			case ' ':
				if (s == NONQUOTE) {
					END_CURRENT_AND_CREATE_NEW;
				} else
					curArg[curFillingIdx++] = commandLine_[idx];
				break;
			case '\"':
				END_CURRENT_AND_CREATE_NEW;
				s = (STATE)((s+1)%2);
				break;
			default:
				curArg[curFillingIdx++] = commandLine_[idx];
			}
			idx++;
		}
		curArg[curFillingIdx++] = L'\0';
		applicationBuf_ = arguments_.front();
	}

};


	
ppi_winapi_prologue(CreateProcessANew, BOOL, 10,
	_In_opt_	LPCSTR, lpApplicationName,
	_Inout_opt_	LPSTR, lpCommandLine,
	_In_opt_	LPSECURITY_ATTRIBUTES, lpProcessAttributes,
	_In_opt_	LPSECURITY_ATTRIBUTES, lpThreadAttributes,
	_In_	BOOL, bInheritHandles,
	_In_	DWORD, dwCreationFlags,
	_In_opt_	LPVOID, lpEnvironment,
	_In_opt_	LPCSTR, lpCurrentDirectory,
	_In_	LPSTARTUPINFOW, lpStartupInfo,
	_Out_	LPPROCESS_INFORMATION, lpProcessInformation);

	PPI_CRITICAL("Not handled in CreateProcessANew: Running application %s %s", lpCommandLine, lpApplicationName);
	BOOL result = TRUE;
		goto call_original_and_return;
ppi_winapi_epilogue(CreateProcessANew, result)

	
const PPI_PRINCIPAL *_getFilePrincipal(const wchar_t *file) {
	PPI_Resource r(file);
	return r.getUntrustedOwner();
}

ppi_winapi_prologue(CreateProcessWNew, BOOL, 10,
	_In_opt_	LPCWSTR, lpApplicationName,
	_Inout_opt_	LPWSTR, lpCommandLine,
	_In_opt_	LPSECURITY_ATTRIBUTES, lpProcessAttributes,
	_In_opt_	LPSECURITY_ATTRIBUTES, lpThreadAttributes,
	_In_	BOOL, bInheritHandles,
	_In_	DWORD, dwCreationFlags,
	_In_opt_	LPVOID, lpEnvironment,
	_In_opt_	LPCWSTR, lpCurrentDirectory,
	_In_	LPSTARTUPINFOW, lpStartupInfo,
	_Out_	LPPROCESS_INFORMATION, lpProcessInformation);

	BOOL result = false;

	if (lpApplicationName != NULL) {

		PPI_CommandLine c(lpApplicationName, lpCommandLine);
		PPI_CRITICAL("PPI_CommandLine returned application as %ls", c.getApplicationPath());
		list<wchar_t *> arg = c.getFileArguments();
		list<const PPI_PRINCIPAL *> arg_principals;
		const PPI_PRINCIPAL *leastTrusted = PPI_PRINCIPAL_MANAGER::getInstance().getCurrentPrincipal();

		if (!arg.empty()) {
			arg_principals.resize(arg.size());
			transform(arg.begin(), arg.end(), arg_principals.begin(), _getFilePrincipal);

			list<const PPI_PRINCIPAL *>::const_iterator pit = arg_principals.begin();

			leastTrusted = arg_principals.front();

			for (list<wchar_t *>::const_iterator it = arg.cbegin(); it != arg.cend(); it++) {
				//PPI_CRITICAL("File Argument list includes %ls, principal: %ls", *it, (*pit)->username);
				leastTrusted = PPI_PRINCIPAL_MANAGER::getInstance().getCurrentPrincipal()->getLessTrusted(leastTrusted, (*pit));
				advance(pit, 1);
			}

		}
				
		if (leastTrusted != PPI_PRINCIPAL_MANAGER::getInstance().getCurrentPrincipal()) {
			PPI_CRITICAL("Executing %ls as %ls Command: %ls", c.getApplicationPath(), leastTrusted->username, lpCommandLine);
			result = OriginalCreateProcessW(L"C:\\WINDOWS\\system32\\runas.exe", c.getRunAsCommand(leastTrusted), lpProcessAttributes, lpThreadAttributes, bInheritHandles,
			dwCreationFlags, lpEnvironment, lpCurrentDirectory, lpStartupInfo, lpProcessInformation);
			goto out;
		}
	}
	goto call_original_and_return;

out:
ppi_winapi_epilogue(CreateProcessWNew, result)