//
// Portable Integrity Protection (PIP) System -
// Copyright (C) 2015 Secure Systems Laboratory, Stony Brook University
//
// This file is part of Portable Integrity Protection (PIP) System.
//
// Portable Integrity Protection (PIP) System is free software: you can
// redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Portable Integrity Protection (PIP) System is distributed in the hope that it
// will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Portable Integrity Protection (PIP) System.  If not, see
// <http://www.gnu.org/licenses/>.
//

#include <stddef.h>
#include "ppi_helperConnection.h"
#include "ppi_calls.h"
#include "ppi_debug.h"
#include "ppi_utils.h"
#include "ppi_level.h"
#include "ppi_redirect.h"
#include "ppi_winternl.h"
#include "ppi_bufferManager.h"

#include <stddef.h>

class PPI_FILE_RENAME_INFORMATION {
public:
	PPI_FILE_RENAME_INFORMATION(const FILE_RENAME_INFORMATION *originalRenameInfo) : _originalRenameInfo(originalRenameInfo) {
		_originalPath = NULL;
		_redirectedRenameInfo = NULL;
	};

	bool isRedirectable() {
		if (getOriginalPath() != NULL)
			return ppi_isRedirectableW(getOriginalPath());
		return false;
	};

	const FILE_RENAME_INFORMATION *getRedirectedStruct() {
		if (_redirectedRenameInfo == NULL) {
			if (!isRedirectable())
				return _originalRenameInfo;
			FILE_RENAME_INFORMATION *t = (FILE_RENAME_INFORMATION *)ppi_bm_path_malloc();
			if (t) {
				t->ReplaceIfExists = _originalRenameInfo->ReplaceIfExists;
				t->RootDirectory = _originalRenameInfo->RootDirectory;

				if (ppi_insertRedirectPrefix(getOriginalPath(), PPI_PRINCIPAL_GET_CURRENT->redirect_prefix, t->FileName, (PPI_BUFFERPOOL_PER_ENTRY_SIZE-offsetof(FILE_RENAME_INFORMATION, FileName))/2) == 0) {
					t->FileNameLength = wcslen(t->FileName)*sizeof(wchar_t);
					if (_redirectedRenameInfo == NULL) {
						_redirectedRenameInfo = t;
						goto out;
					}
				}
				ppi_bm_free(t);
			}
		}
	out:
		return _redirectedRenameInfo;
	};

	const wchar_t *getOriginalPath() {
		if (_originalPath == NULL) {
			wchar_t *path = ppi_bm_path_malloc();
			if (path) {
				memcpy(path, &(_originalRenameInfo->FileName), _originalRenameInfo->FileNameLength);
				path[_originalRenameInfo->FileNameLength/2] = L'\0';
				if (_originalPath == NULL)
					_originalPath = path;
				else
					ppi_bm_free(path);
			}
		}
		return _originalPath;
	}
	~PPI_FILE_RENAME_INFORMATION() {
		if (_originalPath != NULL)
			ppi_bm_free(_originalPath);
		if (_redirectedRenameInfo != NULL)
			ppi_bm_free(_redirectedRenameInfo);
	}
private:
	const FILE_RENAME_INFORMATION *_originalRenameInfo;
	wchar_t *_originalPath;
	FILE_RENAME_INFORMATION *_redirectedRenameInfo;
};



//XXX: need a test case
ppi_winapi_prologue(NtSetInformationFile, NTSTATUS, 5,
  _In_   HANDLE, FileHandle,
  _Out_  PIO_STATUS_BLOCK, IoStatusBlock,
  _In_   PVOID, FileInformation,
  _In_   ULONG, Length,
  _In_   FILE_INFORMATION_CLASS, FileInformationClass
)
	NTSTATUS result;

	if (FileInformationClass == FileRenameInformation) {
		PPI_FILE_RENAME_INFORMATION ppi_ri = PPI_FILE_RENAME_INFORMATION((FILE_RENAME_INFORMATION *)FileInformation);
		if (ppi_ri.isRedirectable() && ppi_ri.getRedirectedStruct()) {
			FileInformation = (void *)ppi_ri.getRedirectedStruct();
			Length = PPI_BUFFERPOOL_PER_ENTRY_SIZE;
		}
		
		result = ppi_callOriginal(NtSetInformationFile, FileHandle, IoStatusBlock, FileInformation, Length, FileInformationClass);
		
		if (result == STATUS_ACCESS_DENIED) {

			switch(FileInformationClass) {
			case FileRenameInformation: {

				// Just ask the helper to rename. But if a benign copy exists, ask helper to rename to shadow storage.

				FILE_RENAME_INFORMATION *fileRenameInformation = (FILE_RENAME_INFORMATION *)FileInformation;

				ppi_pkt_prepare_packets(NtSetInformationFile, pkt, response);
				pkt.FileHandle = FileHandle;
				ppi_pkt_packFileInformation(pkt, FileInformation, FileInformationClass);

				if (ppi_helper_sendRecvCorrectly(pkt, response)) {
					result = response.l_rv;
					*IoStatusBlock = response.IoStatusBlock;
					PPI_INFO("Helper responded with result %Xh", result);
					if (response.l_rv == STATUS_ACCESS_DENIED) {
						PPI_CRITICAL("Helper also failed to open the file with permission error");
//						if (ppi_isRedirectableW(fileName) && !triedRedirectTarget && response.ppi_RetryWithShadow) {
//							PPI_INFO("Helper suggests retry with shadow target");
//							goto try_redirectTarget;
//						}
					} else {
						PPI_INFO("Helper rename gave: status code: %Xh", response.l_rv);
					}
				} else {
					PPI_ERROR("Failed to send and receive from helper correctly");
				}
			
				break;
			}
			default:
				PPI_CRITICAL("Unhandled cases for FileInformationClass!!! %d", FileInformationClass);
			}
		}
	} else
		result = ppi_callOriginal(NtSetInformationFile, FileHandle, IoStatusBlock, FileInformation, Length, FileInformationClass);


ppi_winapi_epilogue(NtSetInformationFile, result)
