//
// Portable Integrity Protection (PIP) System -
// Copyright (C) 2015 Secure Systems Laboratory, Stony Brook University
//
// This file is part of Portable Integrity Protection (PIP) System.
//
// Portable Integrity Protection (PIP) System is free software: you can
// redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Portable Integrity Protection (PIP) System is distributed in the hope that it
// will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Portable Integrity Protection (PIP) System.  If not, see
// <http://www.gnu.org/licenses/>.
//

#include "ppi_helperConnection.h"
#include "ppi_calls.h"
#include "ppi_debug.h"
#include "ppi_utils.h"
#include "ppi_level.h"
#include "ppi_redirect.h"
#include "ppi_regHandleManager.h"


class PPI_QueryDirectoryState {

public:
	PPI_QueryDirectoryState(HANDLE handle): handle(handle) {
		needUpdateCurrentHandle = true;
	};


	void updateCurrentHandle() {
		if (needUpdateCurrentHandle) {
			PPI_DirectoryHandle ppi_dh = PPI_DirectoryHandle(handle);
			const vector<wchar_t *> v = ppi_dh.getAllDirectoryPath();
			vector<wchar_t *>::const_iterator it;
			
			for (it = v.begin(); it != v.end(); it++) {

				if (wcsncmp(*it, ppi_dh.getHandlePath(), 10) == 0) {
					handleList.push_back(handle);
					continue;
				}
				//if (it+1 == v.end())
				//	break;
				HANDLE directoryHandle;


				//struct ppi_tls *tls = ppi_getTls();
				//tls->ppi_isIntercepted = false;
				directoryHandle = CreateFileW(*it, SYNCHRONIZE | FILE_LIST_DIRECTORY | FILE_READ_ATTRIBUTES, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_DIRECTORY | FILE_FLAG_BACKUP_SEMANTICS, NULL);
				//tls->ppi_isIntercepted = true;
				//ppi_call2Hooked(CreateFileWNew, result, FileHandle, DesiredAccess, ObjectAttributes, IoStatusBlock, 0, FILE_ATTRIBUTE_NORMAL, ShareAccess, FILE_OPEN, OpenOptions, NULL, 0);

				//PPI_NEWAPI_LOG("Handle Path: %ls, handle is %d", *it, directoryHandle);
				if (directoryHandle != INVALID_HANDLE_VALUE)
					handleList.push_back(directoryHandle);
				else {
					PPI_ERROR("Failed to get directory handle for path %ls", *it);
				}
			}
			if (v.empty())
				handleList.push_back(handle);

			currentIt = handleList.begin();
			needUpdateCurrentHandle = false;
		}
	}


	NTSTATUS makeQuery(HANDLE Event,
					PIO_APC_ROUTINE ApcRoutine,
					PVOID ApcContext,
					PIO_STATUS_BLOCK IoStatusBlock,
					PVOID FileInformation,
					ULONG Length,
					FILE_INFORMATION_CLASS FileInformationClass,
					BOOLEAN ReturnSingleEntry,
					PUNICODE_STRING FileName,
					BOOLEAN RestartScan) {


		NTSTATUS result = STATUS_NO_MORE_FILES;
		updateCurrentHandle();
		/*
		return ppi_callOriginal(NtQueryDirectoryFile, *currentIt, Event, ApcRoutine, ApcContext, IoStatusBlock,
			FileInformation, Length, FileInformationClass, ReturnSingleEntry, FileName, RestartScan);
*/
		while (currentIt != handleList.end()) {
			result = ppi_callOriginal(NtQueryDirectoryFile, *currentIt, Event, ApcRoutine, ApcContext, IoStatusBlock,
				FileInformation, Length, FileInformationClass, ReturnSingleEntry, FileName, RestartScan);
			if (result == STATUS_NO_MORE_FILES)
				currentIt++;
			else
				break;
		}

		return result;
	}

	~PPI_QueryDirectoryState() {
		vector<HANDLE>::iterator it;
		for (it = handleList.begin(); it != handleList.end(); it++) {
			if (*it != handle)
				CloseHandle(*it);
		}
	}
private:
	const HANDLE handle;
	vector<HANDLE>::iterator currentIt;
	vector<HANDLE> handleList;
	bool needUpdateCurrentHandle;
	//Store a list of the handle, where to start next when the call makeQuery is made
};

class PPI_QueryDirectoryStateManager {

public:
	static PPI_QueryDirectoryStateManager &getInstance() {
		static PPI_QueryDirectoryStateManager instance;
		static bool initialized = false;
		if (!initialized) {
			//instance.initialize();
			initialized = true;
		}
		return instance;
	}

	PPI_QueryDirectoryState *getStateFromHandle(HANDLE handle) {
		unordered_map<HANDLE, PPI_QueryDirectoryState *>::iterator it = handle2State.find(handle);
		if (it == handle2State.end()) {
			handle2State[handle] = new PPI_QueryDirectoryState(handle);
			return handle2State[handle];
		} else
			return it->second;
	}

	void removeState(HANDLE handle) {
		unordered_map<HANDLE, PPI_QueryDirectoryState *>::iterator it = handle2State.find(handle);
		if (it != handle2State.end()) {
			delete it->second;
			handle2State.erase(it);
		}
	}

private:
	PPI_QueryDirectoryStateManager() {};
	PPI_QueryDirectoryStateManager(PPI_QueryDirectoryStateManager const &);
	void operator=(PPI_QueryDirectoryStateManager const &);
	unordered_map<HANDLE, PPI_QueryDirectoryState *> handle2State;
	~PPI_QueryDirectoryStateManager() {
	}

};
	

void ppi_closeFileHandle(HANDLE FileHandle) { 
	PPI_QueryDirectoryStateManager::getInstance().removeState(FileHandle);
}

ppi_winapi_prologue(NtQueryDirectoryFileNew, NTSTATUS, 11,
  _In_      HANDLE, FileHandle,
  _In_opt_  HANDLE, Event,
  _In_opt_  PIO_APC_ROUTINE, ApcRoutine,
  _In_opt_  PVOID, ApcContext,
  _Out_     PIO_STATUS_BLOCK, IoStatusBlock,
  _Out_     PVOID, FileInformation,
  _In_      ULONG, Length,
  _In_      FILE_INFORMATION_CLASS, FileInformationClass,
  _In_      BOOLEAN, ReturnSingleEntry,
  _In_opt_  PUNICODE_STRING, FileName,
  _In_      BOOLEAN, RestartScan
)
	NTSTATUS result;


	if (RestartScan)
		PPI_QueryDirectoryStateManager::getInstance().removeState(FileHandle);

	result = PPI_QueryDirectoryStateManager::getInstance().getStateFromHandle(FileHandle)->makeQuery(Event, ApcRoutine, ApcContext, IoStatusBlock,
		FileInformation, Length, FileInformationClass, ReturnSingleEntry, FileName, RestartScan);
	
	//result = ppi_callOriginal(NtQueryDirectoryFile, FileHandle, Event, ApcRoutine, ApcContext, IoStatusBlock,
		//FileInformation, Length, FileInformationClass, ReturnSingleEntry, FileName, RestartScan);

	if (result == STATUS_NO_MORE_FILES || ReturnSingleEntry)
		PPI_QueryDirectoryStateManager::getInstance().removeState(FileHandle);

out:
ppi_winapi_epilogue(NtQueryDirectoryFileNew, result)

