//
// Portable Integrity Protection (PIP) System -
// Copyright (C) 2015 Secure Systems Laboratory, Stony Brook University
//
// This file is part of Portable Integrity Protection (PIP) System.
//
// Portable Integrity Protection (PIP) System is free software: you can
// redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Portable Integrity Protection (PIP) System is distributed in the hope that it
// will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Portable Integrity Protection (PIP) System.  If not, see
// <http://www.gnu.org/licenses/>.
//

#include <string.h>
#include <uthash.h>
#include "ppi_debug.h"
#include "ppi_regHandleManager.h"
#include "ppi_bufferManager.h"
#include "ppi_redirect.h"

// this is an example of how to do a LRU cache in C using uthash
// http://uthash.sourceforge.net/
// by Jehiah Czebotar 2011 - jehiah@gmail.com
// this code is in the public domain http://unlicense.org/


typedef struct {
	HANDLE regHandle;
} record_key_t;

typedef struct {
	record_key_t key;
	void *path;
	HANDLE rootHandle;
	UT_hash_handle hh;
} record_t;

record_t *cache = NULL;


unsigned int maxCacheSoFar = 0;

void ppi_regHandleManager_handleClosed(HANDLE regHandle)
{
	record_key_t key;
	record_t *entry = NULL;
	
	key.regHandle = regHandle;

	HASH_FIND(hh, cache, &key, sizeof(record_key_t), entry);

	if (entry) {
		HASH_DELETE(hh, cache, entry);
		ppi_bm_free(entry->path);
		ppi_bm_free(entry);
	}
}



void ppi_regHandleManager_saveHandle(HANDLE regHandle, const OBJECT_ATTRIBUTES* oa) {
		
	record_key_t key;
	record_t *entry = NULL;
	
	key.regHandle = regHandle;

	HASH_FIND(hh, cache, &key, sizeof(record_key_t), entry);

	if (entry) {
		//PPI_ERROR("Handle entry already exists!!! Close not intercepted?");
		ppi_regHandleManager_handleClosed(regHandle);
	}

	wchar_t *regName = ppi_bm_path_malloc();
	ppi_utils_extractFilePathFromOA(oa, regName, PPI_PATH_WCHAR_SIZE*sizeof(wchar_t));
	//ppi_regHandleManager_saveHandle(*KeyHandle, ObjectAttributes->RootDirectory, regName);

	entry = (record_t *)ppi_bm_malloc(sizeof(record_t));
	entry->rootHandle = oa->RootDirectory;
	entry->path = regName;
	entry->key = key;
	HASH_ADD(hh, cache, key, sizeof(record_key_t), entry);
	//PPI_INFO("Added handle %x to regcache", regHandle);

	// prune the cache to MAX_CACHE_SIZE
	if (HASH_COUNT(cache) > maxCacheSoFar) {
		PPI_INFO("RegKeyCache max size is now %d", HASH_COUNT(cache));
		maxCacheSoFar = HASH_COUNT(cache);
	}
}

void ppi_regHandleManager_printHandlePath(HANDLE regHandle)
{
	record_key_t key;
	record_t *entry = NULL;
	return;
	
	key.regHandle = regHandle;

	int maxCount = 10;

	PPI_INFO("-------Printing full path...");
	while (maxCount--) {
		HASH_FIND(hh, cache, &key, sizeof(record_key_t), entry);
		if (entry) {
			PPI_INFO("\t Handle: %x: %ls, rootHandle: %x", key.regHandle, cache->path, cache->rootHandle);
			key.regHandle = cache->rootHandle;
		} else
			break;
	}
}


void PPI_FileHandleManager::insertHandle(HANDLE h, const wchar_t *path) {
	unordered_map<HANDLE, wchar_t *>::iterator it = _map.find(h);
	wchar_t *buffer;

	if (it != _map.end())
		closeHandle((*it).first);

	buffer = ppi_bm_path_malloc();
	_bufferCount++;
	if (_bufferCount > _maxCount) {
		PPI_NEWAPI_LOG("Max Buffer usage is now %d", _bufferCount);
		_maxCount = _bufferCount;
	}

	wcscpy_s(buffer, PPI_PATH_WCHAR_SIZE, path);
	_map[h] = buffer;
}


void PPI_FileHandleManager::closeHandle(HANDLE h) {
	unordered_map<HANDLE, wchar_t *>::iterator it = _map.find(h);

	if (it != _map.end()) {
		ppi_bm_free((*it).second);
		_bufferCount--;
		_map.erase(it);
	} else {
		//PPI_NEWAPI_LOG("Closing a handle not registered");
	}
}




const wchar_t *PPI_FileHandleManager::getPathByHandle(HANDLE h) {
	unordered_map<HANDLE, wchar_t *>::iterator it = _map.find(h);

	if (it != _map.end())
		return (*it).second;
	return NULL;
};



#pragma warning(disable: 4996) // wcsnicmp deprecated
#include <winternl.h>

// This makro assures that INVALID_HANDLE_VALUE (0xFFFFFFFF) returns FALSE
#define IsConsoleHandle(h) (((((ULONG_PTR)h) & 0x10000003) == 0x3) ? TRUE : FALSE)
/*
enum OBJECT_INFORMATION_CLASS 
{
    ObjectBasicInformation, 
    ObjectNameInformation,
    ObjectTypeInformation, 
    ObjectAllInformation, 
    ObjectDataInformation
};*/

struct OBJECT_NAME_INFORMATION 
{
    UNICODE_STRING Name; // defined in winternl.h
    WCHAR NameBuffer;
};

typedef NTSTATUS (NTAPI* t_NtQueryObject)(HANDLE Handle, OBJECT_INFORMATION_CLASS Info, PVOID Buffer, ULONG BufferSize, PULONG ReturnLength); 

t_NtQueryObject NtQueryObject()
{
    static t_NtQueryObject f_NtQueryObject = NULL;
    if (!f_NtQueryObject)
    {
        HMODULE h_NtDll = GetModuleHandle(L"Ntdll.dll"); // Ntdll is loaded into EVERY process!
        f_NtQueryObject = (t_NtQueryObject)GetProcAddress(h_NtDll, "NtQueryObject");
    }
    return f_NtQueryObject;
}

DWORD GetNtPathFromHandle(HANDLE h_File, wchar_t* ps_NTPath, size_t size, size_t &copied)
{
    if (h_File == 0 || h_File == INVALID_HANDLE_VALUE)
        return ERROR_INVALID_HANDLE;

    // NtQueryObject() returns STATUS_INVALID_HANDLE for Console handles
    if (IsConsoleHandle(h_File))
    {
		_snwprintf_s(ps_NTPath, size, size, L"\\Device\\Console%04X", (DWORD)(DWORD_PTR)h_File);
        //ps_NTPath->Format(L"\\Device\\Console%04X", (DWORD)(DWORD_PTR)h_File);
        return 0;
    }

    BYTE  u8_Buffer[2000];
    DWORD u32_ReqLength = 0;

    UNICODE_STRING* pk_Info = &((OBJECT_NAME_INFORMATION*)u8_Buffer)->Name;
    pk_Info->Buffer = 0;
    pk_Info->Length = 0;

    // IMPORTANT: The return value from NtQueryObject is bullshit! (driver bug?)
    // - The function may return STATUS_NOT_SUPPORTED although it has successfully written to the buffer.
    // - The function returns STATUS_SUCCESS although h_File == 0xFFFFFFFF
    NtQueryObject()(h_File, (OBJECT_INFORMATION_CLASS)1 /*ObjectNameInformation*/, u8_Buffer, sizeof(u8_Buffer), &u32_ReqLength);

    // On error pk_Info->Buffer is NULL
    if (!pk_Info->Buffer || !pk_Info->Length)
        return ERROR_FILE_NOT_FOUND;

    pk_Info->Buffer[pk_Info->Length /2] = 0; // Length in Bytes!
	copied = pk_Info->Length/2;
	wcscpy_s(ps_NTPath, size, pk_Info->Buffer);

    return 0;
}



bool PPI_RegOA::isRedirectable() const {
	if (printablePath_ == NULL) return false;
	return wcsncmp(printablePath_, L"\\REGISTRY\\USER\\S-", wcslen(L"\\REGISTRY\\USER\\S-")) == 0;
}




OBJECT_ATTRIBUTES *PPI_RegOA::getNextTrustedOA() {

	if (firstReadableOA_ == NULL) {

		if (!isRedirectable())
			goto out;

		const list<const PPI_PRINCIPAL *> *listOfPrincipals = this->principal_->getTrustedPrincipals();

		if (listOfPrincipals->empty())
			goto out;

		UNICODE_STRING *shadowPathUnicode = NULL;
	
		//for (list<const PPI_PRINCIPAL *>::const_iterator it = listOfPrincipals->begin(); it != listOfPrincipals->end(); it++) {
		//shadowPathUnicode = PPI_replaceUserIDinRegAlloc(printablePath_, (*it)->sid);
		
		shadowPathUnicode = PPI_replaceUserIDinRegAlloc(printablePath_, PPI_PRINCIPAL_GET_PRINCIPAL(PPI_MULTIP_BENIGN)->sid);
		if (shadowPathUnicode == NULL)
			goto out;

		PPI_CRITICAL("**** Original Path: %ls, Now Reg path is %ls", printablePath_, shadowPathUnicode->Buffer);

		OBJECT_ATTRIBUTES *ShadowObjectAttributes = (OBJECT_ATTRIBUTES *)ppi_bm_malloc(sizeof(OBJECT_ATTRIBUTES));
		if (ShadowObjectAttributes == NULL) {
			PPI_CRITICAL("Failed to allocate memory for object attribute");
			ppi_deallocUnicodeShadowPath(shadowPathUnicode);
			goto out;
		}

		InitializeObjectAttributes(ShadowObjectAttributes, shadowPathUnicode, original_oa->Attributes, NULL, original_oa->SecurityDescriptor);
		ShadowObjectAttributes->SecurityQualityOfService = original_oa->SecurityQualityOfService;

		firstReadableOA_ = ShadowObjectAttributes;

		//}
	}
out:
	//PPI_NEWAPI_LOG("First readable OA: %p", _firstReadableOA);
	return firstReadableOA_;
}
